"""Logic for loading matrices.
"""

def load_matrices(gen, b_names, fluid_loop, market, subnet, options, op, dt):
    """This function builds constant matrices for multi-time-step optimization
    demands, initial conditions, and utility costs must be updated prior to
    optimization.

    INPUTS:
    gen        -list of components in the plant
    b_names    -list of building names
    fluid_loop -list of cooling tower fluid loops
    subnet     -description of each energy network in the optimizing
    options    -options specifying simulation parameters
    op         -string designation for which efficiency fit curves to use
                (op_mat_a: continuous, op_mat_b: discontinuous but more
                accurate)
    dt         -vector of timestep differences

    OUTPUTS:
    qp         -qp dictionary which includes all the optimization matrices
    """

    n_g = len(gen)
    n_b = len(b_names)
    n_fl = len(fluid_loop)
    n_m = len(market)
    # Count number of lines for each network.
    n_l = sum([len(subnet[net]['line']['node1']) for net in subnet['network_names']])

    qp = {}
    qp['organize'] = {}
    qp['organize']['names'] = [gen[i]['name'] for i in range(n_g)]
    for net in subnet['network_names']:
        qp['organize']['names'].extend([subnet[net]['line']['node1'][i] +'_to_' + subnet[net]['line']['node2'][i] for i in range(len(subnet[net]['line']['node1']))])
    qp['organize']['names'].extend(b_names)
    qp['organize']['names'].extend([fluid_loop[i].name for i in range(n_fl)])

    if dt !=None:
        n_s = len(dt)
        qp['organize']['n_s'] = n_s
        qp['organize']['ic'] = [None for i in range(n_g+n_l+n_b+n_fl)]
    else:
        n_s = 0
    qp['organize']['dispatchable'] = [False for i in range(n_g)]
    qp['organize']['storage_w_penalty'] = [False for i in range(n_g)]
    qp['organize']['enabled'] =  [True for i in range(n_g)]
    qp['organize']['ramp_up'] = [None for i in range(n_g)]
    qp['organize']['ramp_down'] = [None for i in range(n_g)]
    qp['organize']['spin_reserve_states'] = [None for i in range(n_g)]
    qp['organize']['transmission'] = [None for i in range(n_l)]
    qp['organize']['hydro'] = []
    qp['organize']['hydro_dr_line'] = []
    qp['organize']['building'] = {}
    qp['organize']['building']['r'] = [None for i in range(n_b)]
    qp['organize']['building']['electrical_req'] = [None for i in range(n_b)]
    qp['organize']['building']['district_heat_req'] = [None for i in range(n_b)]
    qp['organize']['building']['district_cool_req'] = [None for i in range(n_b)]
    qp['organize']['dt'] = dt
    qp['organize']['market'] = [None for i in range(n_m)]
    qp['organize']['out_vs_state'] = [None for i in range(n_g+n_l+n_b+n_fl)]
    qp['indices'] = [[None] for j in range(n_g+n_l+n_b+n_fl)]
    qp['excess_heat'] = options['excess_heat']
    qp['excess_cool'] = options['excess_cool']
    qp['const_cost'] = [0 for i in range(n_g)]
    qp['organize']['equalities'] = [None for i in range(n_g)]
    qp['organize']['inequalities'] = [None for i in range(n_g)]
    qp['organize']['spin_row'] = [None for i in range(n_g)] 
    qp['organize']['balance'] = {}
    qp['network'] = {}
    qp['const_demand'] = {}
    for net in subnet['network_names']:
        qp['organize']['balance'][net] = [None for i in range(len(subnet[net]['nodes']))]
        nn = len(subnet[net]['nodes'])
        qp['network'][net] = {}
        qp['network'][net]['nodes'] = nn
        qp['network'][net]['node_names'] = [subnet[net]['nodes'][j][0] for j in range(nn)]
        qp['const_demand'][net] = {}
        qp['const_demand'][net]['r_eq'] = [[] for n in range(nn)]
        qp['const_demand'][net]['load'] = [[] for n in range(nn)]
    

    if dt !=None:
        qp, n_ic = count_ic(gen, qp, n_g, n_b, n_l, n_fl) # n_ic is an number, anywhere it is used in an index, requires a -1
    else:
        n_ic = 0

    qp, lb, ub = count_states(gen, subnet, options, n_g, n_b, n_l, fluid_loop, dt, n_ic, op, qp)
    if dt !=None:
        for i in range(n_g+n_l+n_b+n_fl):
            qp['indices'][i].insert(0,[qp['organize']['ic'][i]])
    
    # Expand by number of time steps.
    # Number of states per time step is x_l.
    # Order of states will repeat for n_s timesteps.
    t1s = len(lb)
    if dt !=None:
        total_states = n_s * t1s + n_ic
        for n in range(len(qp['indices'])):
            if not qp['indices'][n][1] is None:
                for t in range(1, n_s):
                    qp['indices'][n].append([k + t1s for k in qp['indices'][n][t]])      
            else:
                qp['indices'][n] = [[None] for t in range(n_s+1)]
        h = [0 for i in range(total_states)]
        qp['f'] = [0 for i in range(total_states)]
        qp['lb'] = [0 for i in range(n_ic)]
        qp['ub'] = [0 for i in range(n_ic)]
        for t in range(n_s):
            qp['lb'].extend(lb)
            qp['ub'].extend(ub)
    else:
        h = [0 for i in range(len(lb))]
        qp['f'] = [0 for i in range(len(lb))]
        qp['lb'] = lb
        qp['ub'] = ub
        
    qp['organize'], n_eq, ec = count_equalities(gen, subnet, options['spin_reserve'], qp['organize'], n_ic, options['end_soc'],dt)
    qp['organize'], n_ineq = count_inequalities(gen, subnet, options['spin_reserve'], qp['organize'], n_b, dt)
    if dt !=None:
        qp['organize'] = expand_organize(qp['organize'],subnet,n_s,n_eq,n_ineq,t1s)
        total_equalities = n_s*n_eq + n_ic + len(ec)
        qp['a_eq'] = [[0 for i in range(total_states)] for j in range(total_equalities)]
        qp['b_eq'] = [0 for j in range(total_equalities)]
        # Initial condition identity matrix.
        for i in range(n_ic):
            qp['a_eq'][i][i] = 1
        total_inequalities = n_s*n_ineq
        qp['a'] = [[0 for i in range(total_states)] for j in range(total_inequalities)]
        qp['b'] = [0 for j in range(total_inequalities)]
    else:
        qp['a_eq'] = [[0 for i in range(t1s)] for j in range(n_eq)]
        qp['b_eq'] = [0 for j in range(n_eq)]
        
        qp['a'] = [[0 for i in range(t1s)] for j in range(n_ineq)]
        qp['b'] = [0 for j in range(n_ineq)]

    # Call sub functions to fill in these matrices now that they are
    # sized.
    if dt !=None:
        qp['a_eq'], qp['b_eq'], h, qp['f'] = generator_equalities(gen, subnet, qp['a_eq'], qp['b_eq'], h, qp['f'], qp['organize'], qp['indices'], dt, op)
        qp['a_eq'], qp['b_eq'] = end_state_constraint(gen, options['end_soc'], qp['a_eq'], qp['b_eq'], qp['indices'], ec)
        qp['a'], qp['b'] = generator_inequalities(gen, qp['a'], qp['b'], qp['organize'], qp['indices'], dt)
        qp['a'], qp['b'] = spin_reserve_inequalities(gen, subnet, options['spin_reserve'], qp['a'], qp['b'], qp['organize'], qp['indices'], dt)
    else:
        qp, h = equality_constraints_step(gen, subnet, qp, h)
        qp['a'] = storage_inequalities_step(gen, qp['a'], qp['organize'], qp['indices'])
        qp['a_eq'] = hydro_equalities_step(gen, subnet, qp['a_eq'], qp['organize'], qp['indices'])
        qp['a_eq'], qp['b_eq'], qp['a'] = spin_reserve_constraints_step(gen, subnet, options['spin_reserve'], qp['a_eq'], qp['b_eq'], qp['a'], qp['organize'], qp['indices'])
    
    qp['a_eq'] = line_equalities(subnet, qp['a_eq'], qp['organize'], qp['indices'], n_s, n_g)
    qp['a'] = line_inequalities(subnet, qp['a'], qp['organize'], qp['indices'], n_s, n_g)
    qp['a_eq'], qp['a'], h, qp['f'], qp['organize'] = building_constraints(b_names,subnet, qp['a_eq'], qp['a'], h, qp['f'], qp['organize'], qp['indices'], dt, n_g, n_l, n_s)
    qp['a_eq'], qp['a'], h, qp['f'], qp['organize'] = fluid_loop_constraints(gen, fluid_loop, subnet, qp['a_eq'], qp['a'], h, qp['f'], qp['organize'], qp['indices'], dt, n_l, n_b, n_s)
    qp['h'] = h
    if op == 'b': #when using fit B record the constant load demands
        qp['const_demand'] = load_const_demand(qp['const_demand'], qp, gen, subnet,  n_s)

    #copy these defaults to refer to later whenever they need to be reset
    qp['nom_vectors'] = {}
    qp['nom_vectors']['h'] = [i for i in qp['h']]
    qp['nom_vectors']['f'] = [i for i in qp['f']]
    qp['nom_vectors']['b'] = [i for i in qp['b']]
    qp['nom_vectors']['b_eq'] = [i for i in qp['b_eq']]
    qp['nom_vectors']['lb'] = [i for i in qp['lb']]
    qp['nom_vectors']['ub'] = [i for i in qp['ub']]
    qp['nom_vectors']['const_cost'] = [i for i in qp['const_cost']]
    return qp

def expand_organize(organize,subnet,n_s,n_eq,n_ineq,t1s):
    #expand out by timestep for disable_generators
    for net in subnet['network_names']:
        for i in range(len(subnet[net]['nodes'])):
            organize['balance'][net][i] = [organize['balance'][net][i] + t*n_eq for t in range(n_s)]
    for n in range(len(organize['heat_vented'])):
        if organize['heat_vented'][n]:
            organize['heat_vented'][n].extend([organize['heat_vented'][n][0] + t*t1s for t in range(1,n_s)])
    for n in range(len(organize['cool_vented'])):
        if organize['cool_vented'][n]:
            organize['cool_vented'][n].extend([organize['cool_vented'][n][0] + t*t1s for t in range(1,n_s)])
    for n in range(len(organize['spin_reserve_states'])):
        if organize['spin_reserve_states'][n]:
            organize['spin_reserve_states'][n] = [organize['spin_reserve_states'][n] + t*t1s for t in range(n_s)]
    if 'spin_reserve' in organize:
        organize['spin_reserve'] = [organize['spin_reserve'] + t*n_ineq for t in range(n_s)]
    for n in range(len(organize['equalities'])):
        if organize['equalities'][n]:
            organize['equalities'][n] = [[eq + t*n_eq for eq in organize['equalities'][n]] for t in range(n_s)]
    for n in range(len(organize['inequalities'])):
        if organize['inequalities'][n]:
            organize['inequalities'][n] = [[ineq + t*n_ineq for ineq in organize['inequalities'][n]] for t in range(n_s)]
    for n in range(len(organize['building']['r'])):
        organize['building']['r'][n] = [organize['building']['r'][n] + t*n_ineq for t in range(n_s)]
    for n in range(len(organize['transmission'])):
        if organize['transmission'][n] is None:
            organize['transmission'][n] = [None for t in range(n_s)]
        else:
            organize['transmission'][n] = [organize['transmission'][n] + t*n_ineq for t in range(n_s)]
    return organize

def count_ic(gen, qp, n_g, n_b, n_l, n_fl):
    """Initial condition for each generator and storage."""
    n_ic = 0
    for i in range(n_g):
        if 'ramp' in gen[i]:
            qp['organize']['ic'][i] = n_ic
            n_ic += 1
        if gen[i]['type'] =='HydroStorage':
            qp['organize']['hydro'].append(i)
            n_ic += 1#second initial condition for the SOC fo the resrvior (1st was for power output)
    for i in range(n_b):#all buildings have an initial temperature state
        qp['organize']['ic'][n_g+n_l+i] = n_ic 
        n_ic += 1#initial condition state
    for i in range(n_fl): #all fluid water loops have an initial temperature state
        qp['organize']['ic'][n_g+n_l+n_b+i] = n_ic 
        n_ic += 1 #initial condition state
    return qp, n_ic


def count_states(gen_all, subnet, options, n_g, n_b, n_l, fluid_loop, dt, x_l, op, qp):
    #organize the order of non-initial condition states
    #states for each generator/storage at t=1
    #spinning reserve for generator/storage at t=1 if option is selected
    #states for each transmission line/pipe at t=1
    #state or power loss in heating and cooling energy balances
    #state for spinning reserve shortfall (target - cumulative of all active generators) 
    #repeat order of generators and lines for t = 2:n_s
    lb = []
    ub = []
    n_fl = len(fluid_loop)
    n_ic = x_l+0
    for i in range(n_g):
        gen = gen_all[i]
        if dt == None and gen['type'] =='HydroStorage':
            s = 1  # Storage treated as generator with 1 state.
            lb.append(gen['x']['lb'])
            ub.append(gen['x']['ub'])
            if 's' in gen_all[i]:  # Spill flow.
                s = 2
                lb.append(gen['s']['lb'])
                ub.append(gen['s']['ub'])
        elif dt == None and gen['type'] in ['Storage','ThermalStorage','ElectricStorage','HydrogenStorage']:
            # Ideal storage is treated as generator with 1 state
            # (additional power output relative to 1st dispatch).
            s = 1
            lb.append(-gen['ramp']['b'][0])
            ub.append(gen['ramp']['b'][1])
            if 'y' in gen:
                # Storage treated as generator with 2 states (additional
                # power output relative to 1st dispatch and charging
                # penalty).
                s = 2
                lb.append(0)
                ub.append((1/gen['stor']['charge_eff'] - gen['stor']['disch_eff']) * gen['ramp']['b'][1])
        elif gen['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator']:
            if op=='b':
                f=1
            else:
                f=0
            states = gen['states'][f]
            s = len(states)
            for k in states:
                lb.append(gen[k]['lb'][f])
                ub.append(gen[k]['ub'][f])
            qp['const_cost'][i] = gen['const_cost']
            if gen['start_cost'] > 0:
                qp['organize']['dispatchable'][i] = True
            elif 'const_demand' in gen and len(list(gen['const_demand'].keys()))>0 and any([gen['const_demand'][n]!=0 for n in gen['const_demand']]):
                qp['organize']['dispatchable'][i] = True
            elif qp['const_cost'][i] > 0 or gen[states[0]]['lb'][1]>0:
                # First state has a non-zero lower bound for
                # second optimization or a non-zero cost at an
                # output of 0.
                qp['organize']['dispatchable'][i] = True
            elif gen['type'] in ['ElectricStorage','ThermalStorage','HydrogenStorage'] and 'y' in gen:
                qp['organize']['storage_w_penalty'][i] = True
        else: #Utilities, Market, Tradepoint, ACDCconverter,...
            states = gen['states'][0]
            s = len(states)
            for k in states:
                lb.append(gen[k]['lb'])
                ub.append(gen[k]['ub'])
                
        # Output is sum of multiple states at each timestep.
        if s>0:
            qp['indices'][i][0] =  [x_l + k for k in range(s)]

        
        if gen['type'] == 'Utility' and gen['source']=='electricity' and s == 2:
            qp['organize']['out_vs_state'][i] = [1,-1]
        elif gen['type'] == 'ACDCConverter':
            # Value is the total transfer from AC to DC.
            qp['organize']['out_vs_state'][i] = [1,-1]
        elif gen['type'] in ['Storage','ThermalStorage','ElectricStorage','HydrogenStorage']:
            # Output state for storage is only state of charge
            # (power in the case of hydro).
            qp['organize']['out_vs_state'][i] = [1]
            qp['organize']['out_vs_state'][i].extend([0 for s in range(len(qp['indices'][i][0])-1)])
        else:
            qp['organize']['out_vs_state'][i] = [1 for k in range(s)]
        x_l += s
        if options['spin_reserve']:
            if gen['type'] == 'HydroStorage':
                lb.append(0)
                ub.append(gen['max_gen_capacity'])
            elif gen['type'] == 'ElectricStorage':
                lb.append(-gen['stor']['peak_charge'])
                ub.append(gen['stor']['peak_disch'])
            elif gen['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'HydrogenGenerator']:
                lb.append(0)
                ub.append(gen['ub'])
            qp['organize']['spin_reserve_states'][i] = x_l
            x_l = x_l+1 

    qp, x_l, lb, ub = count_trans_lines(subnet, qp, x_l, lb, ub, n_g, gen_all)
    if options['spin_reserve'] and any([not qp['organize']['spin_reserve_states'][i] is None for i in range(n_g)]):
        qp['organize']['spin_reserve_states'].append(x_l) #spinning reserve shortfall at t=1
        qp['organize']['spin_reserve_states'].append(x_l+1) #spinning reserve provided as ancillary service at t=1
        x_l += 2
        lb.extend([0,0])
        ub_sr = 0
        for i in range(n_g):
            if qp['organize']['spin_reserve_states'][i] !=None:
                ub_sr += ub[qp['organize']['spin_reserve_states'][i]-n_ic]
        ub.extend([ub_sr,ub_sr])
    qp['organize']['heat_vented'], x_l, lb, ub = vented_energy(gen_all, subnet, options['excess_heat'], x_l, lb, ub, 'h')
    qp['organize']['cool_vented'], x_l, lb, ub = vented_energy(gen_all, subnet, options['excess_cool'], x_l, lb, ub, 'c')
    
    # All buildings have an initial temperature state.
    for i in range(n_b):
        # Output state for building is temperature.
        qp['indices'][n_g+n_l+i][0] = [x_l, x_l+1, x_l+2, x_l+3, x_l+4, x_l+5]
        qp['organize']['out_vs_state'][n_g+n_l+i] = [1,0,0,0,0,0]
        # Add 6 states: temperature, heating, cooling, exceeding upper
        # comfort zone, exceeding lower comfort zone, deviation from last temperature.
        x_l += 6
        # Don't let building below 10 degrees C.
        lb.extend([10, 0, 0, 0, 0, 0])
        # Building stays below 35 degrees C, and excess temperature is
        # less than 10 degrees.
        ub.extend([35, 1e5, 1e5, 10, 10, 10])
    for i in range(n_fl):  #all fluid loops have 1 state
        # qp['organize']['states'][n_g+n_l+n_b+i] = [x_l+1]
        qp['indices'][n_g+n_l+n_b+i][0] = [x_l] #output state for fluid loop is temperature
        qp['organize']['out_vs_state'][n_g+n_l+n_b+i] = [1]
        x_l += 1 # 1 states: Temperature of return water
        lb.append(fluid_loop[i].nominal_return_temperature-fluid_loop[i].temperature_range) #Temperature in C
        ub.append(fluid_loop[i].nominal_return_temperature+fluid_loop[i].temperature_range) #Temperature in C
    return qp, lb, ub


def count_trans_lines(subnet, qp, x_l, lb, ub, n_g, gen):
    # States for transmission lines.
    for net in subnet['network_names']:
        if len(subnet[net]['line']['node1'])>0:
            limit = subnet[net]['line']['limit']
            if net == 'Hydro':
                eff = []
                minimum = subnet[net]['line']['minimum']
            else:
                eff = subnet[net]['line']['eff']
                minimum =[0 for i in range(len(limit))]
            for i in range(len(limit)):
                line = subnet[net]['line']['number'][i]
                qp['indices'][n_g+line][0] = [x_l] #no +1 because index
                if eff == [] or len(eff[i])==1: #uni-directional transfer, or river segment, 1 state for each line
                    qp['indices'][n_g+line][0] = [x_l]
                    lb.append(minimum[i])
                    ub.append(limit[i][0])
                    x_l += 1
                else: #bi-directional transfer, 3 states for each line(state of the line and penalty term in each direction)
                    qp['indices'][n_g+line][0] = [x_l, x_l+1, x_l+2]
                    lb.extend([-limit[i][1],0,0])
                    ub.extend([limit[i][0],limit[i][0]*(1-eff[i][0]), limit[i][1]*(1-eff[i][1])])
                    x_l = x_l+3                  
    return qp, x_l, lb, ub


def vented_energy(generators, subnet, vent, x_l, lb, ub, param):
    """This function adds states at each node to allow purposeful energy
    discharge (venting heat or cooling). Returns a list of vent states.

    Positional arguments:
    generators - (list of Component) List of components in the plant.
    subnet 
    vent - (bool) Whether venting is allowed.
    x_l - (int) Index of last state x.
    lb - (list of float) Lower bound for each state so far.
    ub - (list of float) Upper bound for each state so far.
    param - (str) 'h' or 'c', indicating which type of thermal energy is
        being vented.
    """
    if param == 'h':
        net = 'district_heat'
    elif param == 'c':
        net = 'district_cool'
    gen_names = [generators[i]['name'] for i in range(len(generators))]
    
    if net in subnet['network_names']:
        n_nodes = len(subnet[net]['nodes'])
        energy_vented = [[] for i in range(n_nodes)]
        if vent:
            # Find maximum heat production possible.
            max_gen = 0
            for gen in generators:
                if param in gen['output']:
                    if gen['type'] in ['Storage','ThermalStorage','ElectricStorage','HydrogenStorage']:
                        max_gen = max_gen + gen['stor']['peak_disch']
                    else:
                        if param == 'h' and gen['type'] == 'CombinedHeatPower':
                            max_gen += gen['max_heat']
                        elif param in gen['output'] and gen['output'][param][-1][0] > 0:
                            max_gen += gen['ub']
            # Assume heat can be lost at any node in the network that has a
            # device producing heat.
            
            vent_nodes = 0
            # Matrix for the state associated with venting heat at each
            # district heating node, at each timestep.
            for i_node in range(n_nodes):
                # Identify generators at this node.
                for g_name in subnet[net]['equipment'][i_node]:
                    i_gen = gen_names.index(g_name)
                    if 'h' in generators[i_gen]['output']:
                        # Add single state for heat that is vented to make
                        # energy equality true.
                        energy_vented[i_node].append(x_l + vent_nodes)
                        vent_nodes += 1
                        break
            x_l += vent_nodes
            lb.extend([0 for i in range(vent_nodes)])
            ub.extend([max_gen for i in range(vent_nodes)])
    else:
        energy_vented = []
    return energy_vented, x_l, lb, ub


def count_equalities(gen, subnet, sr, organize, ic, end_soc, dt):
    """Equality constraints:
    initial conditions for each generator and storage device
    electric energy balance at each electric subnet node at t=0, including transmission lines
    heat balance at each district heat subnet node at t=0
    any generator link equalitites (linking states within a generator)
    repeat power balance equalities and link equalities at each step

    INPUTS: 
    gen       -list of generator components
    subnet    -sub network object
    sr        -boolean indicating if you have spinning reserve requirements
    ic        -vector of initial conditions
    end_soc   -desired state of charge for storage at the end of the horizon
    dt        -Time interval of each step

    OUTPUTS:
    organize  -updated qp['organize'] structure
    ec        -end conditions vector
    """

    r_eq = ic #row index of the a_eq matrix and b_eq vector

    #the following puts together the energy balance equations
    # 1 energy/mass balance equation for each subnet node
    # nodes were agregated if their line efficiencies were 1
    for net in subnet['network_names']:
        n = len(subnet[net]['nodes'])
        for i in range(n):
            organize['balance'][net][i] = r_eq
            # Note any demands at this node.
            # organize['demand'][net].append(subnet[net]['load'][i])
            r_eq += 1 # There is an energy/mass balance at this node.

    n_g = len(gen)
    ec = []
    for i in range(n_g):
        #link is a field if there is more than one state and the states are linked by an inequality or an equality
        if 'link' in gen[i] and 'eq' in gen[i]['link']:
            m = len(gen[i]['link']['beq'])  # only hydro has an equality constraint (flow to power conversion)
            organize['equalities'][i] = [r_eq+j for j in range(m)]
            r_eq += m
        if dt is None:
            if gen[i]['type'] == 'HydroStorage':
                organize['hydro'].append(i)
                organize['hydro_dr_line'].append(gen[i]['hydro']['down_river'])
            if 'electrical' in subnet['network_names'] and sr and gen['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'HydrogenGenerator', 'ElectricStorage', 'HydroStorage']:
                organize['spin_row'][i] = r_eq
                r_eq = r_eq+1
        elif not end_soc=='flexible':
            if 'stor' in gen[i]:
                ec.append(i) #special case of final state-of-charge equal to initial state or fixed value
    n_eq = r_eq -ic
    return organize, n_eq, ec
#end of function equality_constraints

def count_inequalities(gen, subnet, sr, organize, n_b, dt):
    ## inequality constraints:
    # 2 ramping constriants for each generator at each timestep (ramp up, ramp down)
    # constraints for each energy storage
    # 2 constraints for each bi-directional transmission line to prevent negative losses
    # repeat all of the above for each time
    #INPUTS:
    #gen        -list of generator components
    #subnet     -sub network
    #sr         -boolean indicating if you include spinning reserve
    #organize   -the organization of matrices from within qp
    #n_b         -integer number of buildings
    #dt          -time interval of each step
    #OUTPUTS:
    #organize   -updated ogranization indices of matrices
    n_g = len(gen)
    r = 0 # row index of the A matrix and b vector
    ramping = [None for i in range(n_g)]
    spin_row = [None for i in range(n_g)]
    #agregate spinning reserve shortfall
    #currently only implemented for electric power
    if 'electrical' in subnet['network_names'] and sr:
        organize['spin_reserve'] = r
        r = r+1 #inequality for net spinning reserve (total of all individual generator spinning reserves)
    if not dt is None:
        #ramping and generator inequalities        
        for i in range(n_g):
            if 'ramp' in gen[i]:
                ramping[i] = r
                r = r+2
            if 'link' in gen[i] and 'ineq' in gen[i]['link']:
                m = len(gen[i]['link']['bineq'])
                organize['inequalities'][i] = [r+j for j in range(m)]
                r += m
            if 'electrical' in subnet['network_names'] and sr and gen['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'HydrogenGenerator', 'ElectricStorage', 'HydroStorage']:
                spin_row[i] = r# constraints, ramping, capacity
                r = r+2
    else: #single timestep matrices
        #charging penalty eqn
        for i in range(n_g):
            if 'stor' in gen[i] and not gen[i]['type'] == 'HydroStorage':
                organize['inequalities'][i] = [r]
                r = r+1
    #transmission line inequalities (penalty terms)
    for net in subnet['network_names']:
        if 'line_eff' in subnet[net]:
            eff = subnet[net]['line']['eff']
            for i in range(len(eff)):
                if len(eff[i]) == 1 or eff[i][1] == 0: #uni-directional transfer, 1 state for each line
                    #do noting, no ineqaulities linking penalty states
                    pass
                else: #bidirectional power transfer
                    organize['transmission'][subnet[net]['line']['number'][i]] = r
                    r = r+2
    
    for i in range (n_b):
        organize['building']['r'][i] = r
        r = r+8 # 8 total ineqaulities per building plus two handled with lb & ub

    #number of generator inequality constraints and energy imbalances at each time step will be r
    # there are 2 ramping constraints on each generator/storage
    for n in range(n_g):
        if not ramping[n] is None:
            organize['ramp_up'][n] = [ramping[n]+t*r  for t in range(len(dt))]
            organize['ramp_down'][n] = [ramping[n]+t*r+1  for t in range(len(dt))]
        if not spin_row[n] is None:
            if dt is None:
                organize['spin_row'][n] = spin_row[n]
            else:
                organize['spin_row'][n] = [spin_row[n]+t*r for t in range(len(dt))]
    return organize, r


def generator_equalities(gen_all, subnet, a_eq, b_eq, h, f, organize, indices, dt, op):
    #this function loads generators/chillers/other components into the appropriate energy balance of a_eq
    #INPUTS:
    #gen_all        -list of generators and components
    #subnet         -the subnetwork you are working with
    #a_eq           -the Aeq matrix
    #b_eq           -the b_eq matrix in Aeq*x = b_eq
    #h              -a diagonal matrix of quadratic costs
    #f              -a matrix of linear costs
    #organize       -organization of constriants
    #dt             -vector of delta time
    #op             -string denoting which optimization fit curves are used ('a' or 'b')
    #OUPUTS:
    #a_eq           -a_eq matrix with equality constraints filled in
    #b_eq           -b_eq matrix with equality constraints filled in
    #h              -diagonal matrix with some cost constraints filled in
    #f              -linear cost matrix with some values filled in
    n_s = len(dt)
    gen_names = [gen_all[i]['name'] for i in range(len(gen_all))]
    for net in subnet['network_names']:
        out = subnet[net]['abbreviation']
        if net == 'hydro':
            a_eq = hydro_equalities(gen_all, subnet, a_eq, organize, indices, dt)
        else:
            for n in range(len(subnet[net]['nodes'])):
                r_eq = organize['balance'][net][n]
                # Only things in equip should have an output or input in this
                # network (electric, heating, cooling).
                equip = subnet[net]['equipment'][n]
                for e_name in equip:
                    k = gen_names.index(e_name)
                    gen = gen_all[k]
                    # link is a field if there is more than one state and the
                    # states are linked by an inequality or an equality.
                    if 'link' in gen and 'eq' in gen['link']:
                        for t in range(n_s):
                            for j in organize['equalities'][k][t]:
                                for z in range(len(gen['link']['eq'][j])):
                                    a_eq[j][indices[k][t+1][z]] = gen['link']['eq'][j][z]
                                b_eq[j] = gen['link']['b_eq'][j]
                    if 'stor' in gen and not gen['type'] == 'HydroStorage':
                        eff = gen['stor']['disch_eff']
                        for t in range(n_s):
                            r_eq = organize['balance'][net][n][t]
                            a_eq[r_eq][indices[k][t+1][0]] = -eff/dt[t]  # State of charge at t.
                            if 'y' in gen['states'][0]:
                                a_eq[r_eq][indices[k][t+1][1]] = -1/dt[t]  # Charging penalty at t.
                            soc_prev = indices[k][t][0]# State of charge at IC.
                            a_eq[r_eq][soc_prev] = eff/dt[t]  # State of charge at t-1.
                    elif not indices[k][1][0] is None:
                        if len(gen['output'][out])>1 and op == 'b':
                            gen_output = gen['output'][out][1]
                        else:
                            gen_output = gen['output'][out][0]

                        if len(gen['states'])>1 and op == 'b':
                            state_names = gen['states'][1]
                        else: 
                            state_names = gen['states'][0]
                        if gen['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator']:
                            fit=0
                            if op=='b':
                                fit=1
                            if len(gen['output'][out])>1:
                                gen_output = gen['output'][out][fit]
                            else:
                                gen_output = gen['output'][out][0]
                            state_names = gen['states'][fit]
                            for j in range(len(state_names)):
                                z = min([j,len(gen_output)-1])
                                for t in range(n_s):
                                    r_eq = organize['balance'][net][n][t]
                                    h[indices[k][t+1][j]] = gen[state_names[j]]['h'][fit]
                                    f[indices[k][t+1][j]] =  gen[state_names[j]]['f'][fit]
                                    a_eq[r_eq][indices[k][t+1][j]] = gen_output[z]
                        else:
                            for t in range(n_s):
                                states = indices[k][t+1]
                                for j in range(len(states)):
                                    z = min([j,len(gen_output)])
                                    r_eq = organize['balance'][net][n][t]
                                    h[states[j]] = gen[gen['states'][0][j]]['h']
                                    f[states[j]] =  gen[gen['states'][0][j]]['f']
                                    a_eq[r_eq][states[j]] = gen_output[z]

                # Any heating loss term to balance equality.
                if net == 'district_heat' and organize['heat_vented'][n]:
                    for t in range(n_s):
                        a_eq[organize['balance'][net][n][t]][organize['heat_vented'][n][t]] = -1
                # Any cooling loss term to balance equality.
                if net == 'district_cool' and organize['cool_vented'][n]:
                    for t in range(n_s):
                        a_eq[organize['balance'][net][n][t]][organize['cool_vented'][n][t]] = -1                
    return a_eq, b_eq, h, f
#end of function generator_equalities

def hydro_equalities(gen, subnet, a_eq, organize, indices, dt):
    """This function loads the mass and energy balances of the hydro network"""
    n_s = len(dt)
    n_g = len(gen)
    downriver = subnet['hydro']['line']['node2'] #node names of the downriver node
    for i in range(n_g):
        if gen[i]['type'] == 'HydroStorage':
            n = gen[i]['subnet_node']['hydro']
            index_dr = gen[i]['hydro']['down_river'] #lines leaving this node, i.e. this node is the upriver node (should be at most 1)
            if subnet['hydro']['nodes'][n] in downriver:
                index_ur = subnet['hydro']['line']['number'][subnet['hydro']['up_river'][n]]
            else:
                index_ur = [] #no upstream nodes
            for t in range(n_s):
                ## Mass balance
                req = organize['balance']['hydro'][n][t]
                a_eq[req][indices[i][t+1][1]] = -12.1/dt[t] #SOC at t in acre-ft / hours converted to ft^3/s:  43560 ft^3/acre-ft * 1 hr / 3600s = 12.1 cfs per acre-ft/hr
                if t==0:
                    a_eq[req][organize['ic'][i]+1] = 12.1/dt[t]#SOC at IC is 1 after dam power IC
                else:
                    a_eq[req][indices[i][t][1]] = 12.1/dt[t]
                #river segments flowing into this node
                for j in range(len(index_ur)):
                    T = subnet['hydro']['line']['time'][subnet['hydro']['up_river'][n][j]]
                    tt = sum(dt[0:t])
                    if tt<=T:                           
                        pass
                        #Do nothing; the inflow rate will be updated in update_matrices
                    elif tt>T and tt<=T+dt[0]:#between initial condition & first step
                        frac = (tt-T)/dt[0] #portion of flow from step 1, remainder from  SourceSink + (1-frac)*Inflow(t=1) : subtracted from beq in update matrices
                        a_eq[req][indices[n_g+index_ur[j]][1]]= frac # Qupriver at step 1, 
                    else:
                        step = 1
                        while tt>(T+sum(dt[0:step])):
                            step = step+1
                        frac = (tt-(T+sum(dt[0:step-1])))/dt[step] #portion of flow from step, remainder from previous step
                        a_eq[req][indices[n_g+index_ur[j]][step]]= frac # Qupriver at t - floor(T)
                        a_eq[req][indices[n_g+index_ur[j]][step-1]]= (1-frac) # Qupriver at t - ceil(T)
                #water flow out of the node
                a_eq[req][indices[n_g+index_dr][t+1]]= -1 #Qdownriver,  
                ## convert power to outflow (other part of balance is done in Generator Equalities when doing the electric network, this is the - Qdownriver
                a_eq[organize['equalities'][i][t]][indices[n_g+index_dr][t+1]]= -1 # Power/(eff*head*84.76) + Spill - Qdownriver = 0
        else:
            pass
            ## add water district here
    return a_eq

def end_state_constraint(all_gen, end_soc, a_eq, b_eq, indices, ec):
    #INPUTS:
    #end_soc        -the state of charge for storage at the last timestep
    #ec             -end condition indices
    #OUTPUTS:
    #a_eq            -equality matrix in Aeq*x = b_eq
    #b_eq            -equality matrix in Aeq*x = b_eq

    n = len(b_eq)-len(ec)
    for j in range(len(ec)):
        i = ec[j]
        if not isinstance(end_soc, str):
            a_eq[n][indices[i][-1]] = 1
            b_eq[n] = end_soc[i]
        elif end_soc=='initial':
            b_eq[n] = 0
            if all_gen[i]['type'] == 'HydroStorage':
                a_eq[n][indices[i][0]+1] = 1
                a_eq[n][indices[i][-1]+1] = -1
            else:
                a_eq[n][indices[i][0]] = 1
                a_eq[n][indices[i][-1]] = -1
        n +=1
    return a_eq, b_eq

def generator_inequalities(gen_all, amat, bmat, organize, indices, dt):
    #this function adds the generator state inequality constraints
    n_g = len(gen_all)
    n_s = len(dt)
    for i in range(n_g):
        gen = gen_all[i]
        #ramping
        if not organize['ramp_up'][i] is None:
            for t in range(n_s):
                up = organize['ramp_up'][i][t]
                down = organize['ramp_down'][i][t]
                #if storage, ramping only affects 1st states
                ramp_states = indices[i][t+1]
                if gen['type'] in ['Storage','ThermalStorage','ElectricStorage','HydrogenStorage']:
                    ramp_states = [ramp_states[0]]
                for s in ramp_states:
                    amat[up][s] = 1/dt[t] #ramp up
                    amat[down][s] = -1/dt[t] #ramp down
                if t==0:  # Ramping from initial condition.
                    amat[up][organize['ic'][i]] = -1/dt[t]  # Ramp up.
                    amat[down][organize['ic'][i]] = 1/dt[t]  # Ramp down.
                else: #condition at previous time step
                    ramp_states0 = indices[i][t]
                    if gen['type'] in ['Storage','ThermalStorage','ElectricStorage','HydrogenStorage']:
                        ramp_states0 = [ramp_states0[0]]
                    for s in ramp_states0:
                        amat[up][s] = -1/dt[t] #ramp up
                        amat[down][s] = 1/dt[t] #ramp down
                bmat[up] = gen['ramp']['b'][0] #ramp up
                bmat[down] = gen['ramp']['b'][1] #ramp down
        #inequality constraints
        if not organize['inequalities'][i] is None:
            
            for t in range(n_s):
                ineq_row = organize['inequalities'][i][t]
                for k in range(len(ineq_row)):
                    if k == 0 and gen_all[i]['type'] in ['ThermalStorage','ElectricStorage'] and 'y' in gen_all[i]['states'][0]:
                        amat[ineq_row[k]][indices[i][t+1][0]] = gen['link']['ineq'][0][0] #soc at t
                        amat[ineq_row[k]][indices[i][t+1][1]] = gen['link']['ineq'][0][1] #charging state at t (-1)
                        if t == 0: #state of charge change from initial condition
                            amat[ineq_row[k]][organize['ic'][i]] = -gen['link']['ineq'][0][0]
                        else:
                            amat[ineq_row[k]][indices[i][t][0]] = -gen['link']['ineq'][0][0] #soc at t-1
                        loss = gen['stor']['self_discharge'] * gen['stor']['usable_size']
                        bmat[ineq_row[k]] = gen['link']['bineq'][k] - loss*(1-gen['stor']['charge_eff'])
                    else:
                        for s in range(len(indices[i][t+1])):
                            amat[ineq_row[k]][indices[i][t+1][s]] = gen['link']['ineq'][k][s] #typically buffer on soc
                        bmat[ineq_row[k]] = gen['link']['bineq'][k]
    return amat, bmat


def spin_reserve_inequalities(gen, subnet, sr, amat, bmat, organize, indices, dt):
    #this function adds spinning reserve inequalities to the A and b matrices (Ax<b)
    #spinning reserve inequalities sum all spinning reserves and individual spinning reserves
    #currently only implemented for electric power
    if 'electrical' in subnet['network_names'] and sr:
        n_g = len(gen)
        n_s = len(dt)
        #total spinning reserve shortfall at each time step
        for t in range(n_s):
            sr_states = [organize['spin_reserve_states'][i][t] for i in range(n_g+1)]
            sr_ancillary = organize['spin_reserve_states'][n_g+1][t]
            for s in sr_states:
                amat[organize['spin_reserve'][t]][s] = -1 #inequality for spinning reserve shortfall: -(shortfall) -sum(sr(i)) + sr_ancillary <= -sr target
            amat[organize['spin_reserve'][t]][sr_ancillary] = 1 #inequality for spinning reserve shortfall: -(shortfall) -sum(sr(i)) +sr_ancillary <= -sr target
        #individual spinning reserve at each time step (limited by ramp rate and peak gen capacity)
        for i in range(n_g):
            if gen[i]['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'HydroStorage', 'HydrogenGenerator']:
                
                for t in range(n_s):
                    states = indices[i][t+1]
                    if gen[i]['type'] == 'HydroStorage':
                        states = states[0]
                    sr_state = organize['spin_reserve_states'][i][t]
                    r = organize['spin_row'][i][t]
                    amat[r][sr_state] = 1/dt[t] #sr+power[t] -power[t-1] <= ramprate*dt
                    for s in states:
                        amat[r][s] = 1/dt[t] 
                    if t==0: #ramping from ic
                        amat[r][organize['ic'][i]] = -1/dt[t] #power at t-1
                    else:
                        states0 = indices[i][t]
                        if gen[i]['type'] == 'HydroStorage':
                            states0 = states0[0]
                        for s in states0:
                            amat[r][s] = -1/dt[t] #power at t-1
                    bmat[r] = gen[i]['ramp']['b'][0] #ramp up constraint
                    amat[r][sr_state] = 1#sr +power <=size
                    amat[r][states] = 1
                    bmat[r] = gen[i]['ub'] #max capacity constraint
        #individual spinning reserve at each time step (limited by discharge rate and storage capacity)
        for i in range(n_g):
            if gen[i]['type'] == 'ElectricStorage':
                for t in range(n_s):
                    sr_state = organize['spin_reserve_states'][i][t] 
                    states = indices[i][t+1]
                    soc = states[0]
                    r = organize['spin_row'][i][t]
                    eff = gen[i]['stor']['dicheff']
                    amat[r][sr_state] = 1 #sr +eff*(soc[t-1] -soc[t])/dt <= peak discharge
                    amat[r][soc] = -eff/dt[t]
                    amat[r+1][sr_state] = 1
                    if t==0: #soc change from ic
                        amat[r][organize['ic'][i]] = eff/dt[t]
                        amat[r+1][organize['ic'][i]] = -eff/dt[t]
                    else:
                        states0 = indices[i][t]
                        soc0 = states0[0]
                        amat[r][soc0] = eff/dt[t] #soc at t-1
                        amat[r+1][soc0] = -eff/dt[t] #soc at t-1
                    bmat[r] = gen[i]['ramp']['b'][2] #peak discharge constraint
    return amat, bmat

def equality_constraints_step(gen, subnet, qp, h):
    #this function loads generators/chiller etc into the appropriate energy balance of a_eq
    organize = qp['organize']
    gen_names = [gen[i]['name'] for i in range(len(gen))]
    for net in subnet['network_names']:
        if not net == 'hydro':
            out = subnet[net]['abbreviation']
            for n in range(len(subnet[net]['nodes'])):
                r_eq = organize['balance'][net][n]
                equip = subnet[net]['equipment'][n]
                for e_name in equip:
                    i = gen_names.index(e_name)
                    gen_i = gen[i]
                    states = qp['indices'][i][0]
                    if 'stor' in gen_i:
                        if out in gen_i['output']:
                            qp['a_eq'][r_eq][states[0]] = gen_i['output'][out][0][0]
                            if not gen_i['type'] == 'HydroStorage'and 'y' in gen_i:# Charging penalty.
                                qp['a_eq'][r_eq][states[0]+1] = -1
                    else:
                        # link is a field for if there is more than one state
                        # and the states are linked by an inequality or an
                        # equality.
                        if 'link' in gen_i and 'eq' in gen_i['link']:
                            req2 = organize['equalities'][i]
                            for j in range(len(req2)):
                                for k in range(len(states)):
                                    qp['a_eq'][req2[j]][states[k]] = gen_i['link']['eq'][j][k]
                                qp['b_eq'][req2[j]] = gen_i['link']['b_eq'][j]
                        if len(gen_i['states']) > 0:
                            if gen_i['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator']:
                                gen_output = gen_i['output'][out][-1]
                                state_names = gen_i['states'][-1]
                                for j in range(len(state_names)):
                                    z = min([j,len(gen_output)-1])
                                    h[states[j]] = gen_i[state_names[j]]['h'][-1]
                                    qp['f'][states[j]] = gen_i[state_names[j]]['f'][-1]
                                    qp['a_eq'][r_eq][states[j]] = gen_output[z]
                            else:
                                gen_output = gen_i['output'][out][-1]
                                state_names = gen_i['states'][-1]
                                for j in range(len(state_names)):
                                    z = min([j,len(gen_output)])
                                    h[states[j]] = gen_i[state_names[j]]['h']
                                    qp['f'][states[j]] = gen_i[state_names[j]]['f']
                                    qp['a_eq'][r_eq][states[j]] = gen_output[z]

                # Any heat loss term to balance equality.
                if net == 'district_heat' and qp['excess_heat']==1 and organize['heat_vented'][n]:
                    qp['a_eq'][r_eq][organize['heat_vented'][n][0]] = -1
                if net == 'district_cool' and qp['excess_cool']==1 and organize['cool_vented'][n]:
                    qp['a_eq'][r_eq][organize['cool_vented'][n][0]] = -1    
    return qp, h

def storage_inequalities_step(gen, amat, organize, indices):
    n_g = len(gen)
    for i in range(n_g):
        if 'stor' in gen[i] and not gen[i]['type'] == 'HydroStorage':
            states = indices[i][0]
            r = organize['inequalities'][i][0]
            amat[r][states[0]] = -(1/gen[i]['stor']['charge_eff'] - gen[i]['stor']['disch_eff'])
            amat[r][states[1]] = -1
    return amat

def hydro_equalities_step(gen, subnet, a_eq, organize, indices):
    if 'hydro' in subnet['network_names']:
        n_g = len(gen)
        upriver = subnet['hydro']['line']['node1'] #node names of the downriver node
        uplines = [i for i in range(len(upriver))]

        #equality: additional power/ (hd*eff*84.67) +spillflow - outflow = nominal power flow (from 1st dispatch)
        #this will allow imposition of instream flow constraints later
        for n in range(len(subnet['hydro']['nodes'])):
            i = subnet['hydro']['equipment'][n]
            index_ur = uplines[upriver.index(subnet['hydro']['nodes'][n])] #lines leaving this node, ie this node is the upriver node (should be at most 1)
            line_out = subnet['hydro']['line']['number'][index_ur]
            if gen[i]['type'] == 'HydroStorage':
                r_eq = organize['equalities'][i]
                # states = organize['states'][i]
                # a_eq[r_eq][states[0]] = gen[i]['stor']['power_2_flow']
                a_eq[r_eq][indices[n_g+line_out][0]] = -1 #qdown river (in excess of original solution from multi-time step)
                # if 's' in gen[i]: #spill flow
                #     a_eq[r_eq][states[1]] = 1
    return a_eq
#end of function hydro_equalities_step

def spin_reserve_constraints_step(gen, subnet, sr, a_eq, b_eq, amat, organize, indices):
    #spinning reserve equalities: individual spinning reserves
    #currently only implemented for electric power
    if 'electrical' in subnet['network_names'] and sr==True:
        n_g = len(gen)
        #individual spinning reserve at each time step (limited by ramp rate and peak gen capacity)
        for i in range(n_g):
            if gen[i]['type'] in ['ElectricGenerator', 'CombinedHeatPower', 'HydroStorage', 'ElectricStorage', 'HydrogenGenerator']:
                states = indices[i][0]
                if gen[i]['type'] == 'HydroStorage':
                    states = states[0] #power output only associated with 1st state
                r_eq = organize['spin_row'][i]
                for s in states:
                    a_eq[r_eq][s] = 1
                b_eq[r_eq] = sum(gen[i]['ub']) #max power
        for i in organize['spin_reserve_states'][:n_g+1]:
            if i != None:
                amat[organize['spin_reserve']][i] = -1#inequality for spinning reserve shortfall: -shortfall - sum(sr(i)) + sr(ancillary) <= -sr target
        amat[organize['spin_reserve']][organize['spin_reserve_states'][n_g+1]] = 1 #inequality for spinning reserve shortfall: -shortfall - sum(sr(i)) + sr(ancillary) <= -sr target
    return a_eq, b_eq, amat
#end of function spin_reserve_constraints_step

def line_equalities(subnet, a_eq, organize, indices, n_s, n_g):
    #this function loads generators/chillers etc into the appropriate energy balance of Aeq
    for net in subnet['network_names']:
        if not net== 'Hydro':
            for k in range(len(subnet[net]['line']['node1'])):
                for n in range(len(subnet[net]['nodes'])):
                    if subnet[net]['line']['node1'][k] in subnet[net]['nodes'][n]:
                        n1 = n
                    if subnet[net]['line']['node2'][k] in subnet[net]['nodes'][n]:
                        n2 = n
                line_num = subnet[net]['line']['number'][k]
                direction = subnet[net]['line']['dir'][k]
                if n_s == 0:
                    line_states = indices[n_g+line_num][0]
                else:
                    line_states = indices[n_g+line_num][1]
                if direction == 'dual':
                    if n_s == 0:
                        line_states = indices[n_g+line_num][0]
                        r_eq = organize['balance'][net][n1]
                        a_eq[r_eq][line_states[0]] = -1 #sending node  (this node, i)-->(connected node), is positive, thus positive transmission is power leaving the node, the penalty from b->a is power not seen at a
                        a_eq[r_eq][line_states[1]] = 0
                        a_eq[r_eq][line_states[2]] = -1
                        r_eq2 = organize['balance'][net][n2]
                        a_eq[r_eq2][line_states[0]] = 1 #receiving node (connected node)-->(this node, i), is positive, thus positive power is power entering the node, the penalty from a->b is power not seen at b                
                        a_eq[r_eq2][line_states[1]] = -1
                        a_eq[+r_eq2][line_states[2]] = 0
                    else:
                        for t in range(n_s):
                            line_states = indices[n_g+line_num][t+1]
                            r_eq = organize['balance'][net][n1][t]
                            a_eq[r_eq][line_states[0]] = -1 #sending node  (this node, i)-->(connected node), is positive, thus positive transmission is power leaving the node, the penalty from b->a is power not seen at a
                            a_eq[r_eq][line_states[1]] = 0
                            a_eq[r_eq][line_states[2]] = -1
                            r_eq2 = organize['balance'][net][n2][t]
                            a_eq[r_eq2][line_states[0]] = 1 #receiving node (connected node)-->(this node, i), is positive, thus positive power is power entering the node, the penalty from a->b is power not seen at b                
                            a_eq[r_eq2][line_states[1]] = -1
                            a_eq[+r_eq2][line_states[2]] = 0
                elif direction == 'forward' or direction == 'reverse':#uni-directional transfer
                    if n_s == 0:
                        line_states = indices[n_g+line_num][0]
                        r_eq = organize['balance'][net][n1]
                        a_eq[r_eq][line_states[0]] = -1
                        r_eq2 = organize['balance'][net][n2]
                        a_eq[r_eq2][line_states[0]] = subnet[net]['line']['eff'][k][0]
                    else:
                        for t in range(n_s):
                            line_states = indices[n_g+line_num][t+1]
                            r_eq = organize['balance'][net][n1][t]
                            a_eq[r_eq][line_states[0]] = -1
                            r_eq2 = organize['balance'][net][n2][t]
                            a_eq[r_eq2][line_states[0]] = subnet[net]['line']['eff'][k][0]                        
    return a_eq
#end of function line_equalities

def line_inequalities(subnet, amat, organize, indices, n_s, n_g):
    #transmission:no connection to previous or later time steps, and no dependence on step size
    for net in subnet['network_names']:
        if 'line' in subnet[net] and not len(subnet[net]['line']['eff']) >0:
            eff = subnet[net]['line']['eff']
            for i in range(len(eff)):
                direction = subnet[net]['line']['dir'][i]
                line = subnet[net]['line']['number'][i]
                if direction == 'dual':
                    for t in range(max([n_s,1])):
                        if n_s == 0:
                            states = indices[n_g+line][0]
                            line_row = organize['transmission'][line]
                        else:
                            states = indices[n_g+line][t+1]
                            line_row = organize['transmission'][line][t]
                        amat[line_row][states[0]] = 1-eff[i][0]
                        amat[line_row][states[1]] = -1
                        amat[line_row][states[2]] = 0 #Pab*(1-efficiency) < penalty a to b
                        amat[line_row+1][states[0]] = -(1-eff[i][1])
                        amat[line_row+1][states[1]] = 0
                        amat[line_row+1][states[2]] = -1 #-Pab*(1-efficiency) < penalty b to a
    return amat


def building_constraints(b_names,subnet, a_eq, amat, h, f, organize, indices, dt, n_g, n_l, n_s):
    #States representing building are: Tzone, Heating, Cooling, Excess T, shortfall T, T deviation
    n_b = len(b_names)
    for i in range(n_b):
        e_n = [n for n in range(len(subnet['electrical']['nodes'])) if i in subnet['electrical']['buildings'][n]][0]
        h_n = [n for n in range(len(subnet['district_heat']['nodes'])) if i in subnet['district_heat']['buildings'][n]][0]
        c_n = [n for n in range(len(subnet['district_cool']['nodes'])) if i in subnet['district_cool']['buildings'][n]][0]
        if n_s>0:
            #electric equality            
            organize['building']['electrical_req'][i] = organize['balance']['electrical'][e_n]
            #heating equality
            organize['building']['district_heat_req'][i] = organize['balance']['district_heat'][h_n]
            for t in range(n_s):
                r_eq = organize['building']['district_heat_req'][i][t]
                a_eq[r_eq][indices[n_g+n_l+i][t+1][1]] = -1 #subtract building heating needs from heating energy balance
            #cooling equality
            organize['building']['district_cool_req'][i] = organize['balance']['district_cool'][c_n]
            for t in range(n_s):
                r_eq = organize['building']['district_cool_req'][i][t]
                a_eq[r_eq][indices[n_g+n_l+i][t+1][2]] = -1 #subtract building cooling needs from cooling energy balance
            for t in range(n_s):
                r = organize['building']['r'][i][t]
                #upper bound inequality
                amat[r+4][indices[n_g+n_l+i][t+1][0]] = 1 #upper buffer >= Ti - (Tset + comfort width/2)
                amat[r+4][indices[n_g+n_l+i][t+1][3]] = -1 #upper buffer >= Ti - (Tset + comfort width/2)
                #lower bound inequality
                amat[r+5][indices[n_g+n_l+i][t+1][0]] = -1 #lower buffer >= (Tset - comfort width/2) -Ti
                amat[r+5][indices[n_g+n_l+i][t+1][4]] = -1 #lower buffer >= (Tset - comfort width/2) -Ti
                #positive deviation inequality
                amat[r+6][indices[n_g+n_l+i][t+1][0]] = 1 #T deviation >= (Ti - Tnominal)
                amat[r+6][indices[n_g+n_l+i][t+1][5]] = -1 #T deviation >= (Ti - Tnominal)
                #negative deviation inequality
                amat[r+7][indices[n_g+n_l+i][t+1][0]] = -1 #T deviation >= (Tnominal- Ti)
                amat[r+7][indices[n_g+n_l+i][t+1][5]] = -1 #T deviation >= (Tnominal- Ti)
        else:
            states = indices[n_g+n_l+i][0]
            #electric equality: put demands into electrical (heating and cooling) balances
            r_eq = organize['balance']['electrical'][e_n]
            organize['building']['electrical_req'][i] = r_eq
            #heating equality
            r_eq = organize['balance']['district_heat'][h_n]
            organize['building']['district_heat_req'][i] = r_eq
            a_eq[r_eq][states[1]] = -1 #subtract building heating needs from heating energy balance
            #cooling equality
            r_eq = organize['balance']['district_cool'][c_n]
            organize['building']['district_cool_req'][i] = r_eq
            a_eq[r_eq][states[2]] = -1 #subtract building cooling needs from cooling energy balance
            r = organize['building']['r'][i]
            #upper bound inequality
            amat[r+4][states[0]] = 1 #upper buffer >= Ti - (Tset+comfort width/2)
            amat[r+4][states[3]] = -1 #upper buffer >= Ti - (Tset + comfort width/2)
            #lower bound inequality
            amat[r+5][states[0]] = -1 #lower buffer >= (Tset - comfort width/2) -Ti
            amat[r+5][states[4]] = -1 #lower buffer >= (Tset - comfort width/2) -Ti
            #positive deviation inequality
            amat[r+6][states[0]] = 1 #T deviation >= (Ti - Tnominal)
            amat[r+6][states[5]] = -1 #T deviation >= (Ti - Tnominal)
            #negative deviation inequality
            amat[r+7][states[0]] = -1 #T deviation >= (Tnominal- Ti)
            amat[r+7][states[5]] = -1 #T deviation >= (Tnominal- Ti)
    return a_eq, amat, h, f, organize


def fluid_loop_constraints(gen, fluid_loop, subnet, a_eq, amat, h, f, organize, indices, dt, n_l, n_b, n_s):
    n_g = len(gen)
    n_fl = len(fluid_loop)
    if n_fl>0:
        equip_names = [gen[i]['name'] for i in range(len(gen))]
    
    for i in range(n_fl):
        #first order temperature model: 0 = -T(k) +T(k-1) +dt/capacitance*(energy balance)
        capacitance = fluid_loop[i].fluid_capacity*fluid_loop[i].fluid_capacitance #water capacity in kg and thermal capacitance in kJ/kgK to get kJ/K
        delta_temperature = fluid_loop[i].nominal_supply_temperature - fluid_loop[i].nominal_return_temperature
        pump_power = fluid_loop[i].pump_power_per_kgs/(delta_temperature*fluid_loop[i].fluid_capacitance) # (kW_e/kgs) / (K *  (kJ_t/kg*K)) = kW_e/kW_t 
        c_list = [equip_names.index(equip) for equip in subnet['cooling_water']['equipment'][i] if gen[equip_names.index(equip)]['type'] == 'Chiller']
        if n_s>0:
            for t in range(n_s):
                r_eq = organize['balance']['cooling_water'][i][t]
                a_eq[r_eq][indices[n_g+n_l+n_b+i][t+1][0]] = -capacitance/(3600*dt[t]) #-T(k)
                a_eq[r_eq][indices[n_g+n_l+n_b+i][t][0]] = capacitance/(3600*dt[t]) #T(k-1)
                for k in c_list:
                    e_r_eq = organize['balance']['electrical'][gen[k]['subnet_node']['electrical']][t] #electric equality
                    for s in indices[k][t+1]:
                        a_eq[e_r_eq][s] -= pump_power #add water pumping load to electric 'cost' of chiller
                        a_eq[r_eq][s] += pump_power #add water pumping heat to fluid loop
        else:
            #single step optimization
            # Temperature state done in update_1step because of dt: # Aeq(r_eq,chiller/ fan states) = (HeatRejected)/Building(i)['Cap*dt
            r_eq = organize['balance']['cooling_water'][i]
            for k in c_list:
                e_r_eq = organize['balance']['electrical'][gen[k]['subnet_node']['electrical']] #electric equality
                for s in indices[k][0]:
                    a_eq[e_r_eq][s] -= pump_power #add water pumping load to electric 'cost' of chiller
                    a_eq[r_eq][s] += pump_power #add water pumping heat to fluid loop
    return a_eq, amat, h, f, organize
#end of function fluid_loop_constraints


def load_const_demand(const_demand, qp, gen, subnet,  n_s):
    """Update demands and storage self-discharge."""
    n_g = len(gen)
    gen_names = [gen[i]['name'] for i in range(n_g)]
    for net in subnet['network_names']:
        for n in range(qp['network'][net]['nodes']): #run through all the nodes in this network
            equip = subnet[net]['equipment'][n]  #equipment at this node
            c_load = [0 for i in range(n_g)]
            for e_name in equip:
                k = gen_names.index(e_name)
                if 'const_demand' in gen[k]  and net in gen[k]['const_demand']:
                    c_load[k] = gen[k]['const_demand'][net]
            if n_s == 0:
                const_demand[net]['r_eq'][n] =  qp['organize']['balance'][net][n] #balance at this node 
                const_demand[net]['load'][n] = c_load 
            else:
                const_demand[net]['r_eq'][n] =  qp['organize']['balance'][net][n]  #balance at this node (t = 1:nS)
                const_demand[net]['load'][n] = [c_load for t in range(n_s)]
    return const_demand


"""Defines logic pertaining to the handling of HDF5 simulation results
files.

Functions:
new_result_file
new_h5_result_filepath
existing_h5_result_filepath
append_dispatch_ic
append_step_result
groom_for_record
write_dispatch_step
write_predicted_step
write_solution
write_data
read_dispatch
extract_dispatch_data
"""


import numpy as np

from eagers.config.path_spec import HDF5_SUFFIX, USER_DIR_SIMRESULTS
from eagers.config.user import DEBUG_MODE
from eagers.basic.gen_limit import chp_heat
from eagers.write.result_description import result_description
from eagers.basic.file_handling import \
    ensure_dirpath, ensure_suffix, find_file_in_userdir
from eagers.basic.hdf5 import h5file_context, DatetimeFloatConverter as DFC


DICT_ACCESS_PROTOCOL = lambda x, p: x.get(p)


def new_result_file(proj_name, names, dimensions, zones, pl):
    """Create a new HDF5 results file.

    Positional arguments:
    proj_name - (str) Name of the project. This becomes the name of the
        file.
    names - (dict of list of str) all generator, storage, node, line, building and fluid_loop names in
        the project.
    dimensions - (dict) Dimension counts for space to be allocated in
        result file structures.
    """
    # Check that the given file name does not already exist.
    filepath = new_h5_result_filepath(proj_name)
    if not DEBUG_MODE and filepath.exists():
        raise ValueError(f'File with name {proj_name!r} already exists.')

    # Open file in "w"rite mode.
    with h5file_context(
            filepath, mode='w', title=f'{proj_name} results') as h5file:
        # Create tables based on specified dimensions.
        disp, pred, solu = result_description(names, dimensions, zones, pl)
        h5file.create_table('/', 'dispatch', disp)
        h5file.create_table('/', 'predicted', pred)
        h5file.create_table('/', 'solution', solu)
      

def result_file_dimensions(preload,plant,date,zones,pl):
    # Create dictionary of element names for results to be saved in.
    names = {}
    names['components'] = [g['name'] for g in preload['gen_qp_form']]
    names['storage'] = [g['name'] for g in preload['gen_qp_form'] if 'stor' in g]
    names['nodes'] = [n['name'] for n in plant['network']]
    names['lines'] = []
    for net in preload['subnet']['network_names']:
        snl = preload['subnet'][net]['line']
        names['lines'].extend([snl['node1'][i]+'_to_'+snl['node2'][i] for i in range(len(snl['node1']))])
    names['buildings'] = [b.name for b in plant['building']]
    names['fluid_loop'] = [fl.name for fl in plant['fluid_loop']]
    if 'district_heat' in preload['subnet']:
        names['heating_nodes'] = [n[0] for n in preload['subnet']['district_heat']['nodes']]
    else:
        names['heating_nodes'] = []
    if 'district_cool' in preload['subnet']:
        names['cooling_nodes'] = [n[0] for n in preload['subnet']['district_cool']['nodes']]
    else:
        names['cooling_nodes'] = []
    names['hydro'] = [g['name'] for g in preload['gen_qp_form'] if g['type'] == 'HydroStorage' ]
    dimensions = dict(
        horizon = len(date)-1,
        zone = max(zones),
        plantloop = max(pl),
        timer = 3,
    )
    return names, dimensions


def new_h5_result_filepath(filename):
    """Generate the path to a new HDF5 results file with the given name.
    """
    return ensure_dirpath(
        USER_DIR_SIMRESULTS / ensure_suffix(filename, HDF5_SUFFIX))


def existing_h5_result_filepath(filename):
    """Generate the path to the existing HDF5 results file with the
    given name.
    """
    return find_file_in_userdir(USER_DIR_SIMRESULTS, filename, HDF5_SUFFIX)


def append_dispatch_ic(proj_name, disp_ic):
    """Append an initial condition result to the file corresponding to
    the given project name.

    Positional arguments:
    proj_name - (str) Name of the project.
    disp_ic - (dict) Data for the initial condition dispatch result.
    """
    # Convert vanilla Python lists to NumPy arrays, and remove empty
    # lists.
    groomed_disp = groom_for_record(disp_ic)
    # Open file in append mode, requiring its existence already.
    filepath = existing_h5_result_filepath(proj_name)
    with h5file_context(filepath, mode='r+') as h5file:
        table = h5file.root.dispatch
        row = table.row
        write_dispatch_step(row, groomed_disp, DICT_ACCESS_PROTOCOL)
        row.append()


def append_step_predicted(proj_name, pred_step):
    """Append a single step prediction to the file corresponding to the
    given project name.

    Positional arguments:
    proj_name - (str) Name of the project.
    pred_step - (dict) Data for this step's predicted result.
    """
    # Convert vanilla Python lists to NumPy arrays, and remove empty lists.
    groomed_pred = groom_for_record(pred_step)
    # Open file in append mode, requiring its existence already.
    filepath = existing_h5_result_filepath(proj_name)
    with h5file_context(filepath, mode='r+') as h5file:
        # Get the table's Row object, write to it, then append it to the
        # table.
        table = h5file.root.predicted
        row = table.row
        write_predicted_step(row, groomed_pred, DICT_ACCESS_PROTOCOL)
        row.append()


def append_step_solution(proj_name, solution):
    """Append a single step solution to the file corresponding to the
    given project name.

    Positional arguments:
    proj_name - (str) Name of the project.
    solution - (dict) Data for this step's solution result.
    """
    # Convert vanilla Python lists to NumPy arrays, and remove empty lists.
    groomed_solu = groom_for_record(solution)
    # Open file in append mode, requiring its existence already.
    filepath = existing_h5_result_filepath(proj_name)
    with h5file_context(filepath, mode='r+') as h5file:
        # Get the table's Row object, write to it, then append it to the
        # table.
        table = h5file.root.solution
        row = table.row
        write_solution(row, groomed_solu, DICT_ACCESS_PROTOCOL)
        row.append()


def append_step_dispatch(proj_name, disp_step):
    """Append a single step result to the file corresponding to the
    given project name.

    Positional arguments:
    proj_name - (str) Name of the project.
    disp_step - (dict) Data for this step's dispatch result.
    """
    # Convert vanilla Python lists to NumPy arrays, and remove empty lists.
    groomed_disp = groom_for_record(disp_step)
    # Open file in append mode, requiring its existence already.
    filepath = existing_h5_result_filepath(proj_name)
    with h5file_context(filepath, mode='r+') as h5file:
        # Get the table's Row object, write to it, then append it to the
        # table.
        table = h5file.root.dispatch
        row = table.row
        write_dispatch_step(row, groomed_disp, DICT_ACCESS_PROTOCOL)
        row.append()


def groom_for_record(data):
    """Convert the given data structure's vanilla Python lists to NumPy
    arrays.

    Positional arguments:
    data - (dict) Data structure.
    """
    # Define function to handle NumPy arrays.
    def groom_ndarray(a):
        if 0 not in a.shape:
            # The array is not empty.
            # If the array can be flattened, flatten it to match the
            # shape of its column description.
            if a.ndim > 1:
                # Number of dimensions > 1.
                n_largedim = np.count_nonzero(np.array(a.shape) > 1)
                if n_largedim <= 1:
                    a = a.flatten()
            return a

    result = {}
    for k, v in data.items():
        if isinstance(v, dict):
            result[k] = groom_for_record(v)
        elif isinstance(v, list):
            if v != []:
                result[k] = groom_ndarray(np.array(v))
        elif isinstance(v, np.ndarray):
            result[k] = groom_ndarray(v)
        elif v != None:
            result[k] = v
    return result


def write_dispatch_step(row, disp_step, access):
    """Write one step of dispatch results to HDF5 file."""
    # Write timestamp.
    row['timestamp'] = DFC.d2f_sgl2sgl(
        access(disp_step, 'timestamp'))
    l = list(disp_step.keys())
    l.remove('timestamp')
    # Write dispatch data.
    for h in l:
        k_list = list(disp_step[h].keys())
        if any([isinstance(disp_step[h][k],dict) for k in k_list]):
            for k in k_list:
                write_data(row, disp_step, access,h+'/'+k+'/')
        else:
            write_data(row, disp_step, access,h+'/')


def write_predicted_step(row, pred_step, access):
    """Write one step of predicted results to HDF5 file."""
    # Write timestamp.
    row['timestamp'] = DFC.d2f_sgl2sgl(
        access(pred_step, 'timestamp'))
    row['cost'] = access(pred_step, 'cost')
    l = list(pred_step.keys())
    l.remove('timestamp')
    l.remove('cost')
    # Write predicted data.
    for h in l:
        k_list = list(pred_step[h].keys())
        if any([isinstance(pred_step[h][k],dict) for k in k_list]):
            for k in k_list:
                write_data(row, pred_step, access,h+'/'+k+'/')
        else:
            write_data(row, pred_step, access,h+'/')


def write_solution(row, solu, access):
    """Write solution results to HDF5 file."""
    row['timestamp'] = DFC.d2f_sgl2sgl(
        access(solu, 'timestamp'))
    row['timer'] = access(solu, 'timer')
    # row['value_heat'] = access(solu, 'value_heat')
    l = list(solu.keys())
    l.remove('timestamp')
    l.remove('timer')
    l.remove('value_heat')
    # Write solution data.
    for h in l:
        k_list = list(solu[h].keys())
        if any([isinstance(solu[h][k],dict) for k in k_list]):
            for k in k_list:
                write_data(row, solu, access,h+'/'+k+'/')
        else:
            write_data(row, solu, access,h+'/')


def write_data(row, source, access, preamble):
    """Generalization for writing data from a given set of fields to the
    corresponding HDF5 table location.

    Positional arguments:
    row - (tables.tableextension.Row) Row iterator to write to.
    source - (object) Object to be read from.
    access - (function) Access protocol. Must accept two arguments. The
        first is the object to be accessed. The second is the name of
        the field to access.
    preamble - (str) String specifying the path within the HDF5 table to
        the write location.
    """
    # Use specified preamble to get to the part of source we care about.
    # The trailing '/' exists for consistency between top-level access,
    # (i.e. preamble = '') and lower-level access, (i.e. preamble =
    # 'x/.../').  This, along with the [:-1] indexing, ensures we end up
    # with an empty list in the top-level case and the correct levels in
    # lower-level cases.
    levels = preamble.split('/')[:-1]
    for lvl in levels:
        source = access(source, lvl)

    # Write data.
    fields = set(source.keys())
    for f in fields:
        to_insert = access(source, f)
        # Use type check instead of equality check to accommodate NumPy
        # arrays, which return a boolean array when compared to another
        # value.
        # Looking for one of these:
        #   - Raw values (e.g. 3, -2.9, 'electric').
        #   - Non-empty NumPy arrays.
        # Shouldn't encounter a non-empty dict or list here.
        if not isinstance(to_insert, type(None)) \
                and (not isinstance(to_insert, (dict, list, np.ndarray))
                    or len(to_insert)):
            row[f'{preamble}{f}'] = to_insert


def read_dispatch(project_name, *args):
    """Return dispatch data for the given project."""
    from eagers.read.excel_interface import ProjectTemplateReader
    from eagers.setup.preload import preload
    proj = ProjectTemplateReader.read_userfile(project_name)
    # ProjectPreload(proj['plant'], proj['test_data'] ,proj['options'])
    proj['preload'] = preload(proj['plant'], proj['test_data'] ,proj['options'])
    components = proj['preload']['gen_qp_form']
    subnet = proj['preload']['subnet']
    filepath = existing_h5_result_filepath(project_name)
    with h5file_context(filepath, 'r') as h5f:
        table = getattr(h5f.root, 'dispatch')
        descr = table.description._v_colobjects
        r_str = dict(components='generator_state', storage = 'storage_state', nodes = 'nodedata', lines = 'line_flow', buildings = 'building', fluid_loop = 'fluid_loop')
        names = {}
        for v in r_str:
            k = r_str[v]
            if k in descr:
                names[v] = descr[k]._v_names
        gen_state = {}
        stor_state = {}
        for net in subnet['network_names']:
            gen_state[net] = {}
            stor_state[net] = {}
            for comp in components:
                k = comp['name']
                if any(k in ne for ne in subnet[net]['equipment']):
                    gen_state[net][k] = table.read(field='generator_state/'+k)
                    if comp['type'] == 'ACDCConverter':
                        if net == 'direct_current':
                            gen_state[net][k] = [j*comp['output']['e'][0][1] if j>0 else j for j in gen_state[net][k]]
                        elif net == 'electrical':
                            gen_state[net][k] = [-j*comp['output']['dc'][0][0] if j<0 else -j for j in gen_state[net][k]]
                    if comp['type'] == 'CombinedHeatPower' and net == 'district_heat':
                        gen_state[net][k] = chp_heat(comp,gen_state[net][k])
                    if k in names['storage']:
                        unusable = comp['stor']['size'] - comp['stor']['usable_size']
                        st_soc = table.read(field='storage_state/'+k)
                        stor_state[net][k] = [j + unusable for j in st_soc]
    return gen_state, stor_state


from copy import copy, deepcopy

from eagers.solver.test_min_cases import test_min_cases
from eagers.basic.all_demands import all_demands, add_min_building_demand, subtract_renewable
from eagers.basic.constrain_min_max import constrain_min_max
from eagers.update.update_matrices_step import update_matrices_step
from eagers.basic.marginal_cost import net_marginal_cost
from eagers.basic.best_eff import best_eff


def dispatch_step(
    gen,
    observer,
    market,
    subnet,
    options,
    one_step,
    date,
    forecast,
    scale_cost,
    dt,
    first_profile,
    v_h,
):
    #creates step by step unit commitment for mcQP method such that N*2^G combinations are checked instead
    # of the full mixed integer approach where 2^(NG) combinations are checked and returns the generator
    #output for each step when optimized individually
    #inputs:
    #gen        - list of dictionaries that describes the components of the plant
    #observer   - current status of all generators, buildings, and fluid loops
    #subnet     - equipment at each node and connections between them
    #options    - user defined options
    #one_step   - instance of OneStep optimization matrices for individual timesteps
    #date       - timestamp for specific timesteps
    #forecast   - forecast for all timesteps
    #scale_cost - cost scale factors for all timesteps 
    #dt         - length of time for all timesteps
    #first_profile  - the dispatch from the previous dispatch horizon

    #OUTPUTS:
    #gen_output  - the solution for generator setpoints at all timesteps in horizon optimized individually
    #stor_power  - the power output of each storage device
    #binary_comb - all of the binary combinations of generators feasible at each time step
    #disp_comb   - the corresponding dispatch for each combination tested
    #cost        - the cost for each combination at each time step
    #verified    - whether a particular combination was tested or not
    #flt         - fluid loop temperature expected at each time step
    #bat         - building average temperature expected at each time step
    #flag        - errors

    # Flag values:
    #   0 -- Standard operation.
    #   1 -- No feasible combinations at one or more time steps.

    flag = 0
    flt = [j for j in observer['fluid_loop_temperature']]
    flc = observer['fluid_loop_capacitance']
    bat = [j for j in observer['building_avg_temp']]

    n_g = len(gen)
    n_s = len(date)-1 #gives number of rows
    gen_output = []
    
    ic = []
    for i in range(n_g):
        if 'Storage' in gen[i]['type']:
            ic.append(observer['stor_state'][i])
        else: 
            ic.append(observer['gen_state'][i])
        gen_output.append([ic[i]]) 
    stor_power = []
        
    i_best = []
    cost = []
    verified = []
    binary_comb = []
    disp_comb = []
    heating = []
    cooling = []
    build_temp = []
    fluid_temp = []
    
    limit = 'initially constrained'
    net_abbrev = [subnet[net]['abbreviation'] for net in subnet['network_names']]
    for t in range(n_s):
        demand = all_demands(forecast, subnet,[t])
        renewable = {}
        for j in range(len(gen)):
            if gen[j]['type'] in ['Solar','Wind']:
                renewable[gen[j]['name']] = forecast['renewable'][j][t]
        step_profile = [first_profile[i][t+1] for i in range(n_g)]
        prev_disp = [gen_output[i][t] for i in range(n_g)]
        sc = [[j[t]] if len(j)>0 else [] for j in scale_cost]
        
        stor_power.append(find_stor_power(gen,prev_disp,step_profile,dt[t]))
        marginal = net_marginal_cost(gen,net_abbrev,market,step_profile,sc,dt[t],any([v_h[n][t] for n in range(len(v_h))]))#update marginal cost
        fb = building_forecast_now(forecast['building'],t)
        min_power, max_power = constrain_min_max(gen, limit, ic, dt[t], sum(dt[0:t+1]))
        update_matrices_step(gen,market,subnet,options,one_step,fb,renewable,demand,sc,marginal,stor_power[t],dt[t],min_power, max_power,step_profile,flt,flc,bat)
        c_red = best_eff(gen,sc,marginal)  
        add_min_building_demand(forecast,subnet,demand,[t])
        subtract_renewable(renewable,subnet,demand)
        v, c, b_c, d_c,ht,cl,bt,ft = test_min_cases(one_step,gen,options,demand,c_red,step_profile,stor_power[t],dt[t])
        verified.append(v)
        cost.append(c)
        binary_comb.append(b_c)
        disp_comb.append(d_c)
        heating.append(ht)
        cooling.append(cl)
        build_temp.append(bt)
        fluid_temp.append(ft)
        if not any(verified[t]) or all([len(k)==0 for k in d_c]):
            flag = 1
            gen_output = first_profile
            print('no feasible combinations to test at step' +  str(t))
            return gen_output, stor_power, binary_comb, disp_comb, build_temp, fluid_temp, cost, verified, flag#ends the function here
        else:
            pass
            # print('Tested ' + str(sum(verified[t])) + ' combinations at step ' + str(t))
        # Update initial conditions and building temperatures.
        i_best.append(min(range(len(cost[t])), key=cost[t].__getitem__))

        best_dispatch = [j for j in d_c[i_best[t]]]
        bat = [j for j in bt[i_best[t]]]
        flt = [j for j in ft[i_best[t]]]
        disp_t = update_stor(gen,best_dispatch,prev_disp,dt[t],limit)#only updates the storage states in IC
        for i in range(n_g):
            gen_output[i].append(disp_t[i])
    return gen_output, stor_power, binary_comb, disp_comb, build_temp, fluid_temp, cost, verified, flag


def update_stor(gen, best_dispatch, prev_disp, dt, limit):
    disp_t = copy(best_dispatch)
    #update storage conditions
    n_g = len(gen)
    for i in range(n_g):
        if 'Storage' in gen[i]['type']:
            if gen[i]['type'] == 'HydroStorage':
                pass
            else:
                loss = (gen[i]['stor']['self_discharge']*gen[i]['stor']['usable_size'])
                if (disp_t[i] + loss) > 0: #discharging
                    d_soc = -(disp_t[i] + loss) * dt * gen[i]['stor']['disch_eff']
                else: #charging
                    d_soc = -(disp_t[i] + loss) * dt * gen[i]['stor']['charge_eff']   
                disp_t[i] = prev_disp[i] +d_soc
                if disp_t[i] < 0:
                    # print('Warning: '+gen[i]['name']+' is going negative by ' + str(disp_t[i]/gen[i]['stor']['usable_size']*100) +' %')
                    disp_t[i] = 0
                elif disp_t[i] > gen[i]['stor']['usable_size']:
                    # print('Warning: '+gen[i]['name']+' is exceeding max discharge by ' + str((disp_t[i]-gen[i]['stor']['usable_size'])/gen[i]['stor']['usable_size']*100) +' %')
                    disp_t[i] = gen[i]['stor']['usable_size']
    return disp_t


def find_stor_power(gen, ic, step_profile, dt):
    #find power from charging or discharging energy storage devices
    n_g = len(gen)
    stor_power =  []
    for i in range(n_g):
        if 'stor' in gen[i]:
            loss = gen[i]['stor']['self_discharge']*gen[i]['stor']['usable_size']
            d_soc = step_profile[i]-ic[i]
            if (d_soc/dt + loss)<0: #discharging
                stor_power.append(-(d_soc/dt + loss)*gen[i]['stor']['disch_eff'])
            else: #charging
                stor_power.append(-(d_soc/dt + loss)/gen[i]['stor']['charge_eff'])
        else:
            stor_power.append(None)
    return stor_power

def building_forecast_now(forecast,t):
    if forecast is None:
        fb = None
    else:
        fb = {}
        # n_b = len(fb['Cap'])
        fb['Cap'] = [i for i in forecast['Cap']]
        f_keys = ['E0','Discomfort','Tmin','Tmax','H2E','C2E','h_min','c_min','deadband','T_avg']
        for f in f_keys:
            fb[f] = copy(forecast[f][t])
        f_keys = ['th_bar','tc_bar','ua_c','ua_h']
        for f in f_keys:
            fb[f] = deepcopy(forecast[f][t])
    return fb



def update_market(gen,market,date):
    market = {}
    # if ~isempty(market)
    #     n_m = length(market.next_closing_time);
    # else
    #     n_m = 0;
    # end
    # for i = 1:1:n_m
    #     if nnz([date(1),date(2)]>=market.next_closing_time(i)) %date(2) is hard coded but should reflect the bidding window
    #         %Right now currently submitting bids 2 hours ahead, so need to
    #         %update every hour to update the awards
    #         if strcmp(market.type{i},'Day-Ahead')
    #             %update closing price
    #             market.closing_price(end+1:end+24,i) =  market_forecast(gen,date); %currently fixed, but should pull from a time series or function
    #             %award bids
    #             if strcmp(market.strategy{i}, 'Price-Taker')
    #                 market.award.time(end+1:end+24,i) = market.bid.time(end-23:end,i);
    #                 market.award.price(end+1:end+24,i) = market.closing_price(end-23:end,i);
    #                 market.award.capacity(end+1:end+24,i) = market.bid.capacity(end-23:end,i);
    #             end
    #         elseif strcmp(market.type{i},'Real-Time')
    #             %update closing price
    #             market.closing_price(end+1,i) =  market_forecast(gen,date(1)); %pulls from a time series 
    #             %award bids
    #             if strcmp(market.strategy{i}, 'Price-Taker')
    #                 [~,new_award] = intersect(round(market.bid.time,5),round(market.next_closing_time,5)); %find the index of the submitted bid
    #                 market.award.time(end+1,i) = market.bid.time(new_award,i);
    #                 market.award.price(end+1,i) = market.closing_price(new_award,i);
    #                 market.award.capacity(end+1,i) = market.bid.capacity(new_award,i);
    #             end
    #         end
                    
    #         %update next closing time
    #         while date(2)>=market.next_closing_time(i)
    #             if strcmp(market.type{i},'Day-Ahead')
    #                 market.next_closing_time(i) = market.next_closing_time(i) + 1;  
    #             elseif strcmp(market.type{i},'Real-Time')
    #                 market.next_closing_time(i) = market.next_closing_time(i) + 1/24; %hard coded for now, need to update later
    return market


"""Interpolate weather."""

from eagers.building.basic.two_way_interp import two_way_interp

import datetime as dt
import copy

def interpolate_weather(weather, d):
    test_weather = {}
    date = []
    test_weather['timestamp'] = date
    for x in range(len(d)):
        if d[x]<weather['timestamp'][0] and d[x].year == weather['timestamp'][0].year:
            date.append(weather['timestamp'][0])
        elif d[x]>weather['timestamp'][-1] and d[x].year == weather['timestamp'][-1].year:
            date.append(weather['timestamp'][-1])
        elif d[x]<weather['timestamp'][0] or d[x]<weather['timestamp'][-1]:
            date.append(dt.datetime(weather['timestamp'][0].year,d[x].month,d[x].day,d[x].hour,d[x].minute,d[x].second))
        else:
            date.append(d[x])
    d0 = weather['timestamp'][0].timestamp()
    def to_timestamp(d,d0):
        return [d[x].timestamp()-d0 for x in range(len(d))]
    for x in weather:
        if x!='timestamp' and x!= 'designation':
            test_weather[x]= two_way_interp(to_timestamp(date,d0),to_timestamp(weather['timestamp'],d0),weather[x])
    return test_weather



def disable_gen(qp, locked):
    # indices is the record of which indicies are associated with each generator
    #locked is the specific combination of generators being tested
    n_g = len(qp['organize']['dispatchable']) #just generators
    qp['x_keep'] = [True for i in range(len(qp['lb']))]
    qp['r_keep'] = [True for i in range(len(qp['a']))]
    qp['req_keep'] = [True for i in range(len(qp['a_eq']))]
    subnet_names = list(qp['const_demand'].keys())
    if not 'n_s' in qp['organize']:
        #single time step optimization
        for i in range(n_g):
            if not qp['organize']['enabled'][i]:
                locked[i] = False
            if not locked[i]:
                # Generator offline; remove states.
                for s in qp['indices'][i][0]:
                    qp['x_keep'][s] = False
                # Remove any associated equality constraints.
                if not qp['organize']['equalities'][i] is None:
                    # Equality constraints.
                    for req in qp['organize']['equalities'][i]:
                        qp['req_keep'][req] = False
                if not qp['organize']['spin_row'][i] is None:
                    # Spinning reserve equality constraint.
                    qp['req_keep'][qp['organize']['spin_row'][i]] = False
                if not qp['organize']['inequalities'][i] is None:
                    # Inequality constraints.
                    for r in qp['organize']['inequalities'][i]:
                        qp['r_keep'][r] = False
        #add in constant electric loads from chillers/pumps in FitB
        qp['b_eq'] = [i for i in qp['updated_b_eq']]#restore values to last time update_matrices_step was called
        qp['lb'] = [i for i in qp['updated_lb']] #might get modified in disable_lb_from_group
        for net in subnet_names:
            if qp['const_demand'][net]['r_eq'] is not None:
                for node in range(qp['network'][net]['nodes']):
                    r_eq = qp['const_demand'][net]['r_eq'][node]
                    qp['b_eq'][r_eq] += sum([qp['const_demand'][net]['load'][node][i] for i in range(n_g) if locked[i]])
    else:
        #multi-timestep optimization
        n_s = qp['organize']['n_s']
        #add in constant electric loads from chillers/pumps in FitB
        if not locked is None:
            for net in subnet_names:
                for node in range(qp['network'][net]['nodes']):
                    for t in range(n_s):
                        r_eq = qp['const_demand'][net]['r_eq'][node][t]
                        qp['b_eq'][r_eq] += sum([qp['const_demand'][net]['load'][node][t][i] for i in range(n_g) if locked[i][t+1]])

        if not locked is None:
            for i in range(n_g):
                if not qp['organize']['enabled'][i]:
                    locked[i] = [False for t in range(len(locked[i]))]
        
            ramp_up = qp['organize']['ramp_up']
            ramp_down = qp['organize']['ramp_down']
            spin_row = qp['organize']['spin_row']
            #deal with initial condition first
            for i in range(n_g):
                if locked[i][0]==False:
                    state = qp['indices'][i][0][0]
                    
                    if locked[i][1]==False or ramp_up[i] is None:#if locked off
                        qp['x_keep'][state] = False
                        qp['req_keep'][qp['organize']['ic'][i]] = False
                    else:
                        #if it turns on right away keep ic = 0 constraint, but change ramping constraints
                        qp['r_keep'][ramp_down[i][0]] = False #eliminate ramp down constraint
                        qp['b'][ramp_up[i][0]] = max([qp['b'][ramp_up[i][0]],-1.001*qp['a'][ramp_up[i][0]][state]*qp['lb'][qp['indices'][i][1][0]]]) #ramp up constraint
                        qp['b_eq'][qp['organize']['ic'][i]] = 0
                for t in range(n_s):
                    if not locked[i][t+1]: #if its offline
                        for s in qp['indices'][i][t+1]:
                            qp['x_keep'][s] = False
                        if (t == n_s-1 and not locked[i][t]) or (not locked[i][t] and not locked[i][t+2]) or qp['organize']['ramp_up'][i] is None:
                            #completely remove all generator states and ramping constraints at this time step if
                            # a) it was off previously & is the last time step, b) was off previously and will remain off at next step, c) has no ramping constraint (shuts off instantly)
                            #remove the ramping constraints
                            s1 = None
                            if not ramp_up[i] is None:
                                qp['r_keep'][ramp_up[i][t]] = False #ramp up constraint
                                qp['r_keep'][ramp_down[i][t]] = False #ramp down constraint
                                if t<n_s-1:
                                    qp['r_keep'][ramp_up[i][t+1]] = False #ramp up constraint
                                    qp['r_keep'][ramp_down[i][t+1]] = False #ramp down constraint
                        else:
                            #keep the first state for ramping constraint to zero
                            s1 = qp['indices'][i][t+1][0]
                            qp['x_keep'][s1] = True
                            #simplify the ramping constraint when ramping to/from offline
                            qp['r_keep'][ramp_up[i][t]] = False #ramp up constraint from previous condition
                            r_down0 = ramp_down[i][t]
                            if not locked[i][t]:
                                qp['r_keep'][r_down0] = False #ramp down constraint
                            else:
                                qp['b'][r_down0] = max([qp['b'][r_down0],-1.001*qp['a'][r_down0][s1]*qp['lb'][s1]]) #ramp down constraint
                            #there is no ramping constraint to next step to remove if it is the last timestep
                            if t==n_s-1:
                                pass
                            #if you are offline in the next step, remove ramp up and ramp down constraint
                            elif not locked[i][t+2]:
                                qp['r_keep'][ramp_up[i][t+1]] = False #ramp up constraint to next time
                                qp['r_keep'][ramp_down[i][t+1]] = False #ramp down constraint to next time
                            else:
                                qp['r_keep'][ramp_down[i][t+1]] = False #amp down constraint to next time
                                r_up2 = ramp_up[i][t+1]
                                qp['b'][r_up2] = max([qp['b'][r_up2],-1.001*qp['a'][r_up2][s1]*qp['lb'][s1]])

                        # Eliminate the spinning reserve state for the individual generator.
                        if not qp['organize']['spin_reserve_states'][i] is None:
                            qp['x_keep'][qp['organize']['spin_reserve_states'][i][t]] = False
                        # Remove any associated equality constraints.
                        if not qp['organize']['equalities'][i] is None:
                            # Equality constraints.
                            qp['req_keep'][qp['organize']['equalities'][i][t]] = False
                        if not s1 is None:
                            qp['ub'][s1] = 0 #Constrained to be off
                            qp['lb'][s1] = 0 #Constrained to be off
                        if not qp['organize']['inequalities'][i] is None:
                            # Inequality constraints.
                            qp['r_keep'][qp['organize']['inequalities'][i][t]] = False
                        if not spin_row[i] is None:
                            # Spinning reserve equality constraint.
                            qp['req_keep'][spin_row[i][t]] = False
        qp['f'] = check_stor_penalty(qp['organize'],qp['indices'],qp['a_eq'],qp['f'],n_g,n_s)

def check_stor_penalty(organize,indices,a_eq,f,n_g,n_s):
    ## small penalty on charging state so that if there are no other generators active at a particular timestep to apply a cost it keeps the charging constraint as an equality
    for i in range(n_g):
        if organize['storage_w_penalty'][i]:
            f_name = list(organize['balance'].keys())
            for k in f_name:
                for n in range(len(organize['balance'][k])):
                    r_eq = organize['balance'][k][n][0]
                    states = indices[i][1]
                    if a_eq[r_eq][states[1]] !=0: #storage is in this energy balance
                        for t in range(n_s):
                            r_eq = organize['balance'][k][n][t]
                            other_gen = [i for i, val in enumerate(a_eq[r_eq]) if val!=0] 
                            if all([f[ind]==0 for ind in other_gen]):
                                f[indices[i][t+1][1]]= 1e-6
    return f


# import numpy as np
from numpy import sin, cos, tan, arcsin, arccos


def solar_calc(longitude, latitude, time_zone, inp_date):
    """ Calculate position of sun and sunrise and sunset times.\n
    Calculated using NOAA solar calculations available at:
    https://www.esrl.noaa.gov/gmd/grad/solcalc/NOAA_Solar_Calculations_day.xls \n
    LONGITUDE is longitude (+ to east).\n
    LATITUDE is latitude (+ to north).\n
    TIME_ZONE is the time zone (+ to east).\n
    INP_DATE is the date as a list of datetime objects, i.e. Jan 1 2017 = 
    datetime(2017,1,1).\n
    SUNRISE and SUNSET are given in fraction of the day, i.e. 6am = 6/24.\n
    AZIMUTH and ZENITH are given in degrees."""

    sunrise = []
    sunset = []
    zenith = []
    azimuth = []
    degrees = 180/3.141592653589793
    radians = 3.141592653589793/180

    tpm = [(x - x.replace(hour=0, minute=0, second=0)).total_seconds() / 24 / 3600  for x in inp_date] # time past midnight [days]
    jd = [x.toordinal() + 1721424.5 - time_zone/24 for x in inp_date]
        # Julian day [days]
        # Julian calendar epoch: Jan 1, 4713 B.C., 12:00:00.0
        # UNIX epoch: Jan 1, 0001 A.D., 00:00:00.0
        # MATLAB epoch: Jan 1, 0000 A.D., 00:00:00.0
        # 1721423.5 days from Julian epoch to UNIX epoch, plus 1 to match
        #     Microsoft Excel's year 1900 exception to the Gregorian calendar.
        # USNO Julian Date Converter:
        #     http://aa.usno.navy.mil/data/docs/JulianDate.php
        # More info on Excel dates:
        #     http://calendars.wikia.com/wiki/Microsoft_Excel_day_number
    for x in range(len(jd)):
        jc = (jd[x] - 2451545) / 36525 # Julian century
        geom_mean_long_sun = (280.46646 + jc * (36000.76983 + jc*0.0003032)) % 360 # [degrees]
        geom_mean_anom_sun = 357.52911 + jc * (35999.05029 - 0.0001537*jc) # [degrees]
        eccent_earth_orbit = 0.016708634 - jc * (0.000042037 + 0.0000001267*jc)
        sun_eq_of_center = float(
            sin(radians * (geom_mean_anom_sun)) * (1.914602 - jc * (0.004817 + 0.000014*jc)) 
            + sin(radians * (2*geom_mean_anom_sun)) * (0.019993 - 0.000101*jc) 
            + sin(radians * (3*geom_mean_anom_sun)) * 0.000289)
        sun_true_long = geom_mean_long_sun + sun_eq_of_center # [degrees]
        mean_obliq_ecliptic = 23 + (26 + (21.448 - jc * (46.815 + jc * (0.00059 - jc*0.001813))) / 60) / 60 # [degrees]
        sun_app_long = sun_true_long - 0.00569 - 0.00478 * sin(radians * (125.04 - 1934.136*jc)) # [degrees]
        obliq_corr = mean_obliq_ecliptic + 0.00256*cos(radians * (125.04 - 1934.136*jc)) # [degrees]
        sun_declin = arcsin(sin(radians * (obliq_corr))* sin(radians * (sun_app_long))) # solar declination [degrees]
        var_y = tan(radians * (obliq_corr/2)) * tan(radians * (obliq_corr/2))
        eq_of_time = 4 * degrees * ( var_y * sin(2*radians * (geom_mean_long_sun)) \
            - 2 * eccent_earth_orbit * sin(radians * (geom_mean_anom_sun)) \
            + 4 * eccent_earth_orbit * var_y * sin(radians * (geom_mean_anom_sun)) * cos(2*radians * (geom_mean_long_sun)) \
            - 0.5 * var_y * var_y * sin(4*radians * (geom_mean_long_sun)) \
            - 1.25 * eccent_earth_orbit * eccent_earth_orbit * sin(2*radians * (geom_mean_anom_sun)))
        ha_sunrise = degrees * (arccos( \
            cos(radians * (90.833)) / (cos(radians * (latitude)) * cos(sun_declin)) \
            - tan(radians * (latitude)) * tan(sun_declin)))
            # sunlight hours [degrees]
        solar_noon = (720 - 4*longitude - eq_of_time + time_zone*60) / 1440
        sunrise.append(solar_noon - ha_sunrise * 4 / 1440) # Local Sidereal Time (LST)
        sunset.append(solar_noon + ha_sunrise * 4 / 1440) # Local Sidereal Time (LST)
        tst = (tpm[x]*1440 + eq_of_time + 4*longitude - 60*time_zone) % 1440 # True Solar Time [min]
        if tst/4 < 0:
            hour_angle = tst/4 + 180 # [degrees]
        else:
            hour_angle = tst/4 - 180 # [degrees]
        zenith.append(degrees * (arccos(sin(radians * (latitude)) * sin(sun_declin) 
            + cos(radians * (latitude)) * cos(sun_declin) * cos(radians * (hour_angle)))))  # [degrees]
        ang = float(degrees * (arccos(((sin(radians * (latitude))*cos(radians * (zenith[x]))) - sin(sun_declin))
            / (cos(radians * (latitude))*sin(radians * (zenith[x]))))))
        if hour_angle>0:
            azimuth.append((ang+180) % 360)#[degrees] (clockwise from N)
        else:
            azimuth.append((540-ang) % 360) #[degrees] (clockwise from N)
    return sunrise, sunset, azimuth, zenith



def constrain_min_max(gen, limit, ic, dt, dt_sum):
    n_g = len(gen)
    
    min_power = []
    max_power = []
    for i in range(n_g):
        upper_bound = 0
        lower_bound = 0
        ramp_down = None
        ramp_up = None
        if 'ramp' in list(gen[i].keys()):
            ramp_down = gen[i]['ramp']['b'][0]
            ramp_up = gen[i]['ramp']['b'][1]
        if gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','CoolingTower','Electrolyzer','HydrogenGenerator'] and len(gen[i]['states'][-1])>0 :
            for j in gen[i]['states'][-1]:
                upper_bound += gen[i][j]['ub'][-1]
        elif gen[i]['type'] == 'HydroStorage':
            pass # Do something to set UB.
        elif gen[i]['type'] == 'Market':
            pass # Do something to set UB.
        elif gen[i]['type'] == 'Utility' and len(gen[i]['states'][-1])>0:
            upper_bound = gen[i][gen[i]['states'][0][0]]['ub']
            if len(gen[i]['states'][-1])>1:
                lower_bound = -gen[i][gen[i]['states'][0][1]]['ub']
        if limit == 'unconstrained' or ic is None or ramp_down is None:
            min_power.append(lower_bound)
            max_power.append(upper_bound)
        else:
            if limit == 'constrained':
                dtr = dt
            else:
                dtr = dt_sum
            min_power.append(max([lower_bound,ic[i]-ramp_down*dtr]))
            max_power.append(min([upper_bound,ic[i]+ramp_up*dtr]))
    return min_power, max_power



import numpy as np

def water_heater(building,T,tank_on,T_exit,T_air,T_source,flow_source,dt,name):
    Cp_water = 4186 #J/kg*K
    den_water = 997 #kg/m^3
    wh = building.water_heater[name]
    wh_gas=0
    wh_elec=0
    e_source = wh['source_side_effectiveness']
    m_cp = wh['tank_volume']*den_water*Cp_water
    T_dif = wh['deadband_temperature_dif']
    if T_dif==None:
        T_dif = 0
    if e_source==None:
        e_source = 1
    if wh['on_cycle_fuel_use']==None:
        q_on_cycle = 0
    else:
        q_on_cycle = wh['on_cycle_fuel_use']*wh['on_cycle_frac2tank']
    if wh['off_cycle_fuel_use']==None:
        q_off_cycle = 0
    else:
        q_off_cycle = wh['off_cycle_fuel_use']*wh['off_cycle_frac2tank']
    UA_on_cycle = wh['on_cycle_loss2ambient']
    UA_off_cycle = wh['off_cycle_loss2ambient']
    q_heater = wh['heater_maximum_capacity']*wh['heater_efficiency']
    a_on = (q_heater + q_on_cycle + UA_on_cycle*T_air + e_source*flow_source*Cp_water*T_source)/m_cp
    b_on = -(UA_on_cycle + e_source*flow_source*Cp_water)/m_cp
    a_off = (q_off_cycle + UA_off_cycle*T_air + e_source*flow_source*Cp_water*T_source)/m_cp
    b_off = -(UA_off_cycle + e_source*flow_source*Cp_water)/m_cp

    sec_on = 0
    t = 0
    while t < dt:
        step = min(60,dt-t) #minutes
        if T<T_exit-T_dif:
            tank_on = True
        if tank_on:
            if b_on == 0:
                T_new = a_on*step + T
            else:
                T_new = (a_on/b_on + T)*float(np.exp(b_on*step)) - a_on/b_on
            if T_new>=T_exit:#turns off partway through step
                tank_on = False
                if b_on == 0:
                    t_on = (T_exit - T)/a_on
                else:
                    t_on = 1/b_on*float(np.log((a_on/b_on + T_exit)/(a_on/b_on + T)))
                sec_on = sec_on + t_on #on for fraction of minute
                #off for remainder of minute
                T = T_exit
                if b_off == 0:
                    T = a_off*(step-sec_on) + T
                else:
                    T = (a_off/b_off + T)*float(np.exp(b_off*(step-sec_on))) - a_off/b_off
            else:
                sec_on = sec_on + step
        else:
            if b_off == 0:
                T = a_off*step + T
            else:
                T = (a_off/b_off + T)*float(np.exp(b_off*step)) - a_off/b_off
        t += step

    if wh['heater_fuel']=='naturalgas':
        wh_gas += sec_on/dt*wh['heater_maximum_capacity'] #Watts
    else:
        wh_elec += sec_on/dt*wh['heater_maximum_capacity'] #Watts
    if q_on_cycle>0:
        if wh['on_cycle_fuel_type']=='naturalgas':
            wh_gas += sec_on/dt*wh['on_cycle_fuel_use'] #Watts
        else:
            wh_elec +=sec_on/dt*wh['on_cycle_fuel_use'] #Watts
    if q_off_cycle>0:
        if wh['off_cycle_fuel_type']=='naturalgas':
            wh_gas += (dt-sec_on)/dt*wh['off_cycle_fuel_use'] #Watts
        else:
            wh_elec += (dt-sec_on)/dt*wh['off_cycle_fuel_use'] #Watts
    return T,tank_on,wh_gas,wh_elec


from eagers.building.basic.eval_curve import eval_curve

def flow_resolver_supply(building,supply_temperature,load,flow,T_return,method,on_loop):
    ##supply side
    ##Pump at start of supply side
    supply_flow = [0 for x in range(len(building.plant_supply_nodes['name']))]
    for j in on_loop:
        in_n = building.plant_supply_equip['inlet'][j] 
        out_n = building.plant_supply_equip['outlet'][j] 
        if building.plant_supply_equip['type'][j]=='mixer':
            if type(in_n) != list:
                supply_flow[out_n] = supply_flow[in_n]
            else:
                supply_flow[out_n] = sum([supply_flow[k] for k in in_n]) 
        elif building.plant_supply_equip['type'][j]=='splitter':
            dist,supply_temperature = load_distribution(building,supply_temperature,out_n,supply_flow[in_n],T_return,load,method) 
            if type(out_n) != list:
                supply_flow[out_n] = dist[0]
            else:
                for i in range(len(out_n)):
                    supply_flow[out_n[i]] = dist[i]
        elif building.plant_supply_equip['type'][j]=='pump:variablespeed' or building.plant_supply_equip['type'][j]=='pump:constantspeed':
            supply_flow[out_n] = flow 
        else:
            supply_flow[out_n] = supply_flow[in_n] 
    return supply_flow, supply_temperature

def load_distribution(building,supply_temperature,out,flow_in,T_in,net_load,method):
    ## need equipment on the loop to add up to the load
    ## equipment + bypass flows = flow in
    #not doing anything if equipment is in series on a supply branch (not sure if this is ever a problem)
    names = []
    bypass = None
    for j in range(len(building.plant_supply_equip['name'])):
        for k in range(len(out)):
            if (type(building.plant_supply_equip['inlet'][j]) == list and out[k] == building.plant_supply_equip['inlet'][j][0]) or out[k] == building.plant_supply_equip['inlet'][j]:
                names.append(building.plant_supply_equip['name'][j]) 
                if building.plant_supply_equip['bypass'][j]:
                    bypass = len(names) 
                    break
    n = len(names) 
    flow_out = [0 for i in range(n)]
    #first distribution of flow
    if method =='optimal':
        if bypass==None or bypass !=0:
            flow_out[0] = flow_in 
        else:
            flow_out[1] = flow_in 
    elif method =='uniformload':
        #split evenly amongst non-bypass branches
        if bypass == None: #no bypass branch
            flow_out = [flow_in/n for i in range(n)]
        else:
            flow_out = [flow_in/(n-1) for i in range(n)]
            flow_out[bypass-1] = 0 
    elif method =='sequentialload':
        if bypass==None or bypass !=1:
            k=0
        else:
            k=1
        flow_out[k] = flow_in 
    elif method =='uniformplr':
        print('need work in load_distribution of UniformPLR')
    elif method =='sequentialuniformplr':
        print('need work in load_distribution of SequentialUniformPLR')
    else:
        print('need work in load_distribution of manage_plant_equip')

    ## Re-distribution of flow
    error_load = 1 
    tol = 1e-2 
    if net_load>0:
        while abs(error_load)>tol:
            supply_temperature,PLR,load,flow_max,PLR_opt = PLR_load(building,supply_temperature,names,flow_out,T_in,net_load) 
            flow_out = min(flow_max,flow_out) 
            if load == None: #cooling tower, cant calculate load yet
                error_load = 0 
            else:
                error_load = (net_load - sum(load))/net_load 
                sys_with_spare_cap = [i for i in range(len(load)) if load[i]>0 and flow_out[i]<flow_max[i]] 
                if error_load>0 and len(sys_with_spare_cap)==0:
                    error_load = 0 
            if method=='optimal':
                if error_load>tol and any([PLR[i]<PLR_opt[i] for i in range(len(PLR))]):
                    #increase load up to PLR_opt in order of equipment 
                    for j in range(n):
                        if PLR[j]<PLR_opt[j]:
                            flow_out[j] = flow_out[j]*min(PLR_opt[j]/PLR[j],(load[j]+error_load*net_load)/load[j]) 
                            break 
                elif error_load<-tol and any([True for i in range(len(PLR)) if PLR[i]>PLR_opt[i]]):
                    #decrease load to PLR_opt
                    for j in range(n):
                        if PLR[j]>PLR_opt[j]:
                            flow_out[j] = flow_out[j]*min(PLR_opt[j]/PLR[j],(load[j]+error_load*net_load)/load[j]) 
                            break 
                elif abs(error_load)>tol:
                    #distrubte error in load uniformly
                    sys_with_spare_cap = [i for i in range(len(load)) if load[i]>0 and flow_out[i]<flow_max[i]] 
                    for j in sys_with_spare_cap:
                        flow_out[j] = flow_out[j]*(load[j] + (net_load - sum(load))/len(sys_with_spare_cap))/load[j]
            elif method=='uniformload':
                for j in sys_with_spare_cap:
                    flow_out[j] = flow_out[j]*(load[j] + (net_load - sum(load))/len(sys_with_spare_cap))/load[j] 
                flow_out = [f*flow_in/sum(flow_out) for f in flow_out]  #normalize to correct total mass flow
                error_load = 0 
            elif method=='sequentialload':
                if error_load>tol:
                #increase load up to max PLR in order of equipment 
                    if flow_out[k]<flow_max[k]:
                        flow_out[k] = min(flow_max[k],flow_out[k]*(load[k]+error_load*net_load)/load[k]) 
                    else:
                        k += 1 
                        if k == bypass-1:
                            k += 1 
                        if k<=n:
                            flow_out[k] = 0.5*flow_max[k] 
                elif error_load<-tol:
                    #decrease load 
                    flow_out[k] = max(0,flow_out[k]*(load[k]+error_load*net_load)/load[k]) 
                    if flow_out[k] == 0:
                        k -= 1 
                        if k == bypass-1:
                            k -= 1 
            elif method=='uniformplr':
                print('need work in load_distribution of UniformPLR')
            elif method=='sequentialuniformplr':
                print('need work in load_distribution of SequentialUniformPLR')
            else:
                print('need work in load_distribution of manage_plant_equip')
            if bypass!=None:
                flow_out[bypass-1] = max(0,flow_in - (sum(flow_out)- flow_out[bypass-1])) 
    return flow_out, supply_temperature


def PLR_load(building,supply_temperature,names,flow,T_in,net_load):
    Cp_water = 4186  #J/kg*K
    n = len(names) 
    PLR = [0 for i in range(n)]
    load = [flow[i]*net_load/sum(flow) for i in range(n)]
    flow_max = [0 for i in range(n)]
    PLR_opt = [0 for i in range(n)]
    for j in range(n):
        k = building.plant_supply_equip['name'].index(names[j])
        den = building.plant_loop['fluid_density'][building.plant_supply_equip['loop'][k]]
        if (building.plant_supply_equip['type'][k]=='chiller:electric:reformulatedeir' or
            building.plant_supply_equip['type'][k]=='chiller:electric:eir' or
            building.plant_supply_equip['type'][k]=='chiller:constantcop'):
            T_out = T_in - load[j]/(flow[j]*Cp_water)
            PLR[j],load[j],T_out,Q_avail,c = chiller_PLR(building,names[j],flow[j],T_in,T_out)  
            if PLR[j]*Q_avail/(Cp_water*(T_in - T_out))<flow[j]:
                flow_max[j] = PLR[j]*Q_avail/(Cp_water*(T_in - T_out)) 
            else:
                flow_max[j] = building.chiller[c]['chilled_water_flow_ref']*den #max flow in kg/s
            PLR_opt[j] = building.chiller[c]['opt_part_load']
            out = building.plant_supply_equip['outlet'][building.plant_supply_equip['name'].index(names[j])]
            supply_temperature[out] = T_out
        elif building.plant_supply_equip['type'][k]=='boiler:hotwater':
            T_out = T_in + load[j]/(flow[j]*Cp_water)
            PLR[j],load[j],T_out,b = boiler_PLR(building,names[j],flow[j],T_in,T_out) 
            if load[j]/(Cp_water*(T_out - T_in))<flow[j]:
                flow_max[j] = load[j]/(Cp_water*(T_out - T_in)) 
            else:
                flow_max[j] = building.boiler[b]['design_flow_rate']*den #max flow in kg/s
            PLR_opt[j] = building.boiler[b]['optimal_part_load']
            out = building.plant_supply_equip['outlet'][building.plant_supply_equip['name'].index(names[j])]
            supply_temperature[out] = T_out
        elif building.plant_supply_equip['type'][k]=='coolingtower:singlespeed':
            load = [] 
            PLR[j] = 1 
            PLR_opt[j] = 1 
            flow_max[j] = building.cooling_tower[names[j]]['water_flow']*den
        elif building.plant_supply_equip['type'][k]=='waterheater:mixed':
            load = [] 
            PLR[j] = 1 
            PLR_opt[j] = 1 
            flow_max[j] = building.plant_loop['max_flow_rate'][building.plant_supply_equip['loop'][k]]*den
        elif (building.plant_supply_equip['type'][k]=='chiller:enginedriven' or
              building.plant_supply_equip['type'][k]=='chiller:absorption' or 
              building.plant_supply_equip['type'][k]=='chiller:absorption:indirect' or 
              building.plant_supply_equip['type'][k]=='chiller:combustionturbine' or
              building.plant_supply_equip['type'][k]=='chillerheater:absorption:directfired' or
              building.plant_supply_equip['type'][k]=='chillerheater:absorption:doubleeffect' or
              building.plant_supply_equip['type'][k]=='coolingtower:twospeed' or
              building.plant_supply_equip['type'][k]=='coolingtower:variablespeed:merkel'):
            print('need examples of these types of equipment__' + building.plant_supply_equip['type'][k])
        elif  building.plant_supply_equip['bypass'][k]:
            flow_max[j] = sum(flow) 
    return supply_temperature,PLR,load,flow_max,PLR_opt

def boiler_PLR(building,b,flow,T_in,T_out):
    Cp_water = 4186  #J/kg*K
    load = flow*Cp_water*(T_out- T_in) 
    PLR = min(load/building.boiler[b]['capacity'],building.boiler[b]['max_part_load']) 
    if load>0 and PLR<building.boiler[b]['min_part_load']:
        PLR = building.boiler[b]['min_part_load']
    load = PLR*building.boiler[b]['capacity']
    return PLR,load,T_out,b

def chiller_PLR(building,c,flow,T_in,T_out):
    ##see pg 763 about standard IPLV values
    Cp_water = 4186  #J/kg*K
    Q_evap = flow*Cp_water*(T_in - T_out) # (Watts) kg/s*J/kgK * K = J/s
    #need to update with actual condenser water outlet temperature
    CCFT = eval_curve(building.curves,building.chiller[c]['capacity_temperature_curve'],[T_out,building.chiller[c]['condenser_water_temperature_ref']]) 
    Q_avail = building.chiller[c]['capacity']*CCFT #(Watts)
    if flow>0:
        min_T_out = T_in - Q_avail/(Cp_water*flow) 
        T_out = max(T_out,min_T_out) 
        PLR = min(Q_evap/Q_avail,building.chiller[c]['max_part_load']) 
    else: 
        PLR = 0
    load = PLR*Q_avail
    return PLR,load,T_out,Q_avail,c


"""Complimentary Quadratic Programming method (not mcQP method).

Functions:
cqp_method
rule_one
rule_two
rule_three
rule_four
"""

from eagers.solver.ecos_qp import ecos_qp
from eagers.solver.sort_solution import sort_solution, sort_eh
from eagers.update.disable_gen import disable_gen
from eagers.basic.hdf5 import DatetimeFloatConverter as DFC


def cqp_method(names, gen, qp, first_disp, date):
    """Positional arguments:
    gen - (list of dict) Generator dictionary reperesentations.
    qp - (dict)
    first_disp - (2D list) First dispatch.
    date - (list of datetime)
    """
    n_s = len(date) - 1
    n_g = len(gen)
    locked = [[True for t in range(n_s+1)] for i in range(n_g)]
    for i in range(n_g):
        if not gen[i]['enabled']:
            locked[i] = [False for t in range(n_s+1)]
    lower_bound = [0 for i in range(n_g)]
    upper_bound = [0 for i in range(n_g)]
    dx = [[0 for t in range(n_s)] for i in range(n_g)]
    dt_seconds = [(date[t+1] - date[t]).total_seconds() for t in range(n_s)]
    dt = [dt_seconds[t]/3600 for t in range(n_s)]
    for i in range(n_g):
        if qp['organize']['dispatchable'][i]:
            states = gen[i]['states'][-1]
            for j in range(len(states)):
                lower_bound[i] += gen[i][states[j]]['lb'][-1]
                upper_bound[i] += gen[i][states[j]]['ub'][-1]
            # Default to off when initial dispatch is below lb.
            for t in range(n_s):
                if first_disp[i][t+1]<lower_bound[i]:
                    locked[i][t] = False
        if 'ramp' in gen[i]:
            dx[i] = [dt[t]*gen[i]['ramp'] for t in range(n_s)]
    # Make sure it can shut down in time from initial condition.
    for i in range(n_g):
        if qp['organize']['dispatchable'][i]:
            if first_disp[i][0] > 0 and not all(locked[i]):
                d = first_disp[i][0]
                t = 0
                while d > 0:
                    r = qp['organize']['ramp_down'][t]
                    d -= qp['b'][r]
                    if d > 0 and not locked[i][t]:
                        locked[i][t] = True
                    t += 1

    feasible = 1
    attempt = 0
    lb_relax = 1

    # attempt: Integer value describing the number of attempts before
    # reaching feasibility. This determines how close components must be to
    # their lower bound from below to be considered online.
    # 
    # n: Represents the percent of lower bounds on your first try. Just use
    # the locked matrix given, then do unit commitment based on
    # optimal_state > lb * perc_lb
    perc_lb = [0.9, 0.75, 0.5, 0.2, 0.1, 0, -1]
    while feasible != 0 and attempt < len(perc_lb):
        if attempt > 0:
            # Not the first try. Lower limit for online threshold.
            lb_relax = perc_lb[attempt-1]
            # Only change label for unit commitment gens, and don't change the
            # label for initial conditions.
            for i in range(n_g):
                if qp['organize']['dispatchable'][i]:
                    # Default to on unless offline in initial dispatch.
                    locked[i] = [True if first_disp[i][t] > (lower_bound[i] * lb_relax) else False for t in range(len(first_disp[i]))]
        disable_gen(qp, locked)
        x, feasible = ecos_qp(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'], qp['ub'], qp['lb'], qp['x_keep'], qp['r_keep'], qp['req_keep'])
        attempt += 1


    if feasible == 0:
        v_h = [[True if j<1 else False for j in s_eh] for s_eh in sort_eh(x, qp)] #do you value heat generation, or is there tons of extra
        solution = sort_solution(x, qp, names, gen, date, v_h)

        # ----------------------------------------------------------------
        # The following section of code implements superfluous heuristics.
        # ----------------------------------------------------------------
        # 
        # first_disp = solution['dispatch']
        # cost = sum(net_cost(gen,solution.dispatch,date))
        # for i in range(n_g):
        #     if qp.organize.dispatchable[i]:
        #         locked, cost, solution, starts = rule_one(
        #             gen, locked, cost, solution, date, qp, dx, first_disp, i)
        #         locked, cost, solution, stops = rule_two(
        #             gen, locked, cost, solution, date, starts, qp, dx,
        #             first_disp, i)
        #         locked, cost, solution = rule_three(
        #             gen, locked, cost, solution, date, starts, stops, qp, dx,
        #             first_disp, i)
        # _, _, solution = rule_four(
        #     gen, locked, cost, solution, date, qp, dx, upper_bound)
    solution['lb_relax'] = [lb_relax]
    
    return solution


def rule_one(gen, locked, cost, solution, date, qp, dx, first_disp, i):
    # function [locked,cost,solution,starts] = rule_one(gen,locked,cost,solution,date,qp,dx,first_disp,i)
    # %% Rule 1: turn off for longer time at start if possible
    # n_s = length(dx(:,1));
    # index = (1:n_s)';
    # starts = nonzeros(index.*((locked(2:end,i)-locked(1:n_s,i))>0)); % if on at t = 3, start = 3, where IC is t=0
    # if ~locked(1,i) && ~isempty(starts)
    #     p = 0;
    #     ramp_up = dx(starts(1),i);
    #     while ramp_up<first_disp(starts(1)+1,i) && (starts(1)-p>0)
    #         ramp_up = ramp_up+dx(starts(1)-p,i);
    #         p = p+1;
    #     end
    #     if starts(1)-p>0
    #         l2 = locked;
    #         l2(1:(starts(1)-p+1),i) = false;
    #         disable_gen(qp,l2);%Disable generators here
    #         [x,Feasible] = call_solver(qp);
    #         if Feasible == 1
    #             sol_new = sort_solution(x,qp);
    #             new_cost = sum(net_cost(gen,sol_new.dispatch,date));
    #             if new_cost<cost
    #                 locked = l2;
    #                 cost = new_cost;
    #                 solution = sol_new;
    #             end
    #         end
    #     end
    #     if length(starts)>1
    #         starts = starts(2:end);
    #     else
    #         starts = [];
    #     end
    # end
    # end%ends function rule_one

    locked, cost, solution, starts = 0, 0, 0, 0  # TODO: Translate from Matlab.
    return locked, cost, solution, starts


def rule_two(
        gen, locked, cost, solution, date, starts, qp, dx, first_disp, i):
    # function [locked,cost,solution,stops] = rule_two(gen,locked,cost,solution,date,starts,qp,dx,first_disp,i)
    # %% Rule 2: If off for a long enough segment in first dispatch, try turning off for as much of that as possible given ramp rates
    # n_s = length(dx(:,1));
    # index = (1:n_s)';
    # stops = nonzeros(index.*((locked(1:n_s,i)-locked(2:end,i))>0)); % if off at t = 12, stop = 12, where IC is t=0
    # for k = 1:1:length(starts)
    #     if ~isempty(stops) && length(stops)>=k
    #         if sum(dx(stops(k):starts(k)-1,i))>(first_disp(stops(k),i) + first_disp(starts(k)+1,i)) %can ramp all the way down, and back up, and be off for 1 step
    #             l2 = locked;
    #             %find step when it can hit zero given setting at Disp(stops(k)
    #             n=1;
    #             ramp_down = dx(stops(k),i);
    #             while ramp_down<first_disp(stops(k),i)
    #                 ramp_down = ramp_down+dx(stops(k)+n,i);
    #                 n = n+1;
    #             end
    #             p = 1;
    #             ramp_up = dx(starts(k),i);
    #             while ramp_up<first_disp(starts(k)+1,i)
    #                 ramp_up = ramp_up+dx(starts(k)-p,i);
    #                 p = p+1;
    #             end
    #             l2((stops(k)+n):(starts(k)-p+1),i) = false;
    #             disable_gen(qp,l2);%Disable generators here
    #             [x,feasible] = call_solver(qp);
    #             if feasible == 1
    #                 sol_new = sort_solution(x,qp);
    #                 new_cost = sum(net_cost(gen,sol_new.dispatch,date));
    #                 if new_cost<cost
    #                     locked = l2;
    #                     cost = new_cost;
    #                     solution = sol_new;
    #                 end
    #             end
    #         end
    #     end
    # end
    # end%ends function rule_two

    locked, cost, solution, stops = 0, 0, 0, 0  # TODO: Translate from Matlab.
    return locked, cost, solution, stops


def rule_three(
        gen, locked, cost, solution, date, starts, stops, qp, dx, first_disp,
        i):
    # function [locked,cost,solution] = rule_three(gen,locked,cost,solution,date,starts,stops,qp,dx,first_disp,i)
    # %% Rule 3: try turning off @ end
    # n_s = length(dx(:,1));
    # if length(stops)>length(starts)
    #     n=stops(end);
    #     ramp_down = dx(n,i);
    #     while ramp_down<first_disp(stops(end),i) && n<(n_s) && n>0
    #         ramp_down = ramp_down+dx(n,i);
    #         n = n-1;
    #     end
    #     if n<(n_s)
    #         l2 = locked;
    #         l2((n+1):n_s+1,i) = false;
    #         disable_gen(qp,l2);%Disable generators here
    #         [x,feasible] = call_solver(qp);
    #         if feasible == 1
    #             sol_new = sort_solution(x,qp);
    #             new_cost = sum(net_cost(gen,sol_new.dispatch,date));
    #             if new_cost<cost
    #                 locked = l2;
    #                 cost = new_cost;
    #                 solution = sol_new;
    #             end
    #         end
    #     end
    # end
    # end%ends rule_three

    locked, cost, solution = 0, 0, 0  # TODO: Translate from Matlab.
    return locked, cost, solution


def rule_four(gen, locked, cost, solution, date, qp, dx, upper_bound):
    # function [locked,cost,solution] = rule_four(gen,locked,cost,solution,date,qp,dx,upper_bound)
    # %% Rule 4: if on for a short time and sufficient capacity in remaining active generators, turn gen off completely
    # n_s = length(dx(:,1));
    # n_g = length(gen);
    # index = (1:n_s)';
    # for i = 1:1:n_g
    #     starts = nonzeros(index.*((locked(2:end,i)-locked(1:n_s,i))>0)); % if on at t = 3, start = 3, where IC is t=0
    #     stops = nonzeros(index.*((locked(1:n_s,i)-locked(2:end,i))>0)); % if off at t = 12, stop = 12, where IC is t=0
    #     if qp.Organize.Dispatchable(i)
    #         if locked(1,i)
    #             if length(stops)>1
    #                 stops = stops(2:end);
    #             else
    #                 stops = [];
    #             end
    #         end
    #         for k = 1:1:length(stops)
    #             if sum(dx(starts(k):stops(k),i))<upper_bound(i) && sum(locked(:,i))<floor(n_s/4)%can only ramp to 1/2 power and less than 1/4 of horizon
    #                 l2 = locked;
    #                 l2(starts(k):(stops(k)+1),i)= false;
    #                 disable_gen(qp,l2);%Disable generators here
    #                 [x,feasible] = call_solver(qp);
    #                 if feasible == 1
    #                     sol_new = sort_solution(x,qp);
    #                     new_cost = sum(net_cost(gen,sol_new.dispatch,date));
    #                     if new_cost<cost
    #                         locked = l2;
    #                         cost = new_cost;
    #                         solution = sol_new;
    #                     end
    #                 end
    #             end
    #         end
    #     end
    # end
    # end%ends function rule_four

    locked, cost, solution = 0, 0, 0  # TODO: Translate from Matlab.
    return locked, cost, solution



"""

Functions:
test_cases
disable_lb_from_group
cost_estimate
"""

from eagers.update.disable_gen import disable_gen
from eagers.solver.ecos_qp import ecos_qp
from eagers.solver.sort_solution import sort_gen_disp_step, sort_building_step, sort_eh
from eagers.solver.case_marginal import case_marginal


def test_cases(qp,gen,combinations,n_gen,c_red,stor_power,p_groups):
    n_g = len(gen)
    n_b = len(qp['organize']['building']['r'])
    n_l = len(qp['organize']['transmission'])
    n_fl = 0
    if 'cooling_water' in qp['organize']['balance']:
        n_fl = len(qp['organize']['balance']['cooling_water'])
    n_c = len(combinations)
    verified = [False for i in range(n_c)]
    cost = [float('nan') for i in range(n_c)]
    disp_comb = [[] for i in range(n_c)]
    heating = [[0 for j in range(n_b)] for i in range(n_c)]
    cooling =  [[0 for j in range(n_b)] for i in range(n_c)]
    build_temp = [[0 for j in range(n_b)] for i in range(n_c)]
    fluid_temp = [[0 for j in range(n_fl)] for i in range(n_c)]

    stor_abbrev = []
    stor_index = []
    for i in range(n_g):
        if 'stor' in gen[i]:
            f = list(gen[i]['output'].keys())
            if not f[0] in stor_abbrev:
                stor_abbrev.append(f[0])
            stor_index.append(i)
    if not stor_power is None:
        stor_offset = stor_power
    else:
        stor_offset = [0 for i in range(n_g)]

    test_more = True
    best_cost = float('inf')
    k = 0
    n = n_gen[0]
    while test_more:
        if cost[k]!=cost[k] or cost[k] < best_cost:
            disable_gen(qp, combinations[k])
            if not p_groups is None:
                for type_gen in p_groups:
                    if len(p_groups[type_gen])>0 and len(p_groups[type_gen][0])>1:
                        disable_lb_from_group(qp,combinations[k],p_groups[type_gen]) #If generators are grouped, disable the lb on generators 2-n in the most expensive active group
            x, flag1 = ecos_qp(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'], qp['ub'], qp['lb'], qp['x_keep'], qp['r_keep'], qp['req_keep'])
            if flag1 == 0:
                cost[k] = sum([qp['const_cost'][i] for i in range(n_g) if combinations[k][i]])
                cost[k] += sum([0.5 * x[i]* qp['h'][i]*x[i] + x[i]*qp['f'][i] for i in range(len(x))])
                if cost[k] < best_cost:
                    best_cost = cost[k]
                sol_disp = sort_gen_disp_step(x, qp)
                stor_disch = [0 for i in range(n_g)]
                stor_pen = [0 for i in range(n_g)]
                sol_build = sort_building_step(x,qp)
                for si in stor_index:
                    s = qp['indices'][si][0][0]
                    stor_disch[si] = x[s]
                    stor_pen[si] = qp['h'][s]
                    sol_disp[si] += stor_offset[si]
                disp_comb[k] = sol_disp
                heating[k] = sol_build['heating']
                cooling[k] = sol_build['cooling']
                build_temp[k] = [x[qp['indices'][n_g+n_l+i][0][0]] for i in range(n_b)]
                fluid_temp[k] = [x[qp['indices'][n_g+n_l+n_b+i][0][0]] for i in range(n_fl)]
                eh = sort_eh(x, qp)
                if len(eh) == 0 or any([eh[i]>1e-1 for i in range(len(eh)) if isinstance(eh[i],(float,int))]):
                    v_h = False
                else:
                    v_h = True
                marginal = case_marginal(sol_disp,gen,qp,v_h)
                cost = cost_estimate(c_red,gen,cost,verified,marginal,stor_disch,stor_pen,stor_abbrev,combinations,k)
            else:
                cost[k] = float('inf')
            verified[k] = True
            

        test_more = False
        if k<n_c-1 and (n_gen[k+1]==n or n_gen[k+1]<=n_gen[0]+1):
            test_more = True
        elif k<n_c-1: #check if any possible cases with more generators could save money
            j = 1
            while k+j<n_c:
                if cost[k+j] != cost[k+j] or cost[k+j]<best_cost:
                    test_more = True
                    break
                j+=1
            n= n_gen[k+1]
        k += 1
        if all([j== float('inf') for j in cost]):
            print('no feasible solutions')
    return verified, cost,  disp_comb, heating, cooling,build_temp,fluid_temp

def disable_lb_from_group(qp,combination,groups):
    #If generators are grouped, disable the lb on generators 2-n in the most expensive active group
    if groups!=None:
        k = -1
        #find last group that has an active gen
        while k+1<len(groups) and any([j in combination for j in groups[k+1]]):
            k += 1
        if k>=0:
            #set lb on generators 2 through n to 0
            g = groups[k]
            for i in range(1,len(g)):
                states = qp['indices'][g[i]][0]
                qp['lb'][states[0]] = 0


def cost_estimate(c_red,gen,cost,verified,marginal,stor_disch,stor_pen,stor_abbrev,combinations,k):
    """Estimate the best achievable cost of an additional generator by adding
    its cost per kW at peak efficiency and subtracting the current marginal
    cost times the new generator's capacity at peak efficiency.
    """
    n_g = len(gen)
    d_cost = [0 for i in range(n_g)]
    stor_offset = {}
    stor_used = {}
    stor_cost = {}
    stor_avg_pen = {}
    for j in stor_abbrev:
        stor_used[j] = 0
        stor_cost[j] = 0
        stor_offset[j] = [0 for i in range(n_g)]
    for j in range(n_g):
        if c_red['p'][j]>0:
            d_cost[j] = - marginal[c_red['ab'][j]]*c_red['p'][j] + c_red['p'][j]*c_red['c'][j]/c_red['eff'][j]
        if stor_disch[j]>0:
            abbrev = list(gen[j]['output'].keys())[0]
            for i in range(n_g):
                if c_red['p'][i]>0:
                    if abbrev=='dc' and 'e' in gen[i]['output'] and gen[i]['output']['e'][0][0]>0:
                        abbrev2 = 'e'
                    elif abbrev=='e' and 'dc' in gen[i]['output'] and gen[i]['output']['dc'][0][0]>0:
                        abbrev2 = 'dc'
                    else:
                        abbrev2 = abbrev
                    if abbrev2 in gen[i]['output'] and gen[i]['output'][abbrev2][0][0]>0: #penalty can only be reduced by adding generators (of the correct type) if it was dawing power from the storage
                        gen_ub = gen[i]['ub']*gen[i]['output'][abbrev2][0][0]
                        stor_offset[abbrev][i] = min(stor_disch[j],gen_ub-stor_offset[abbrev][i]) #best case reduction in storage use

            stor_used[abbrev] += stor_disch[j]
            stor_cost[abbrev] += 0.5*stor_disch[j]**2*stor_pen[j]
    for abbrev in stor_abbrev:
        if stor_used[abbrev] == 0:
            stor_avg_pen[abbrev] = 0
        else:
            stor_avg_pen[abbrev] = stor_cost[abbrev]/(.5*stor_used[abbrev]**2)

    on_test = [i for i in range(len(combinations[k])) if combinations[k][i]]
    for i in range(len(combinations)):
        if not verified[i] and (cost[i] != cost[i] or cost[i]<cost[k]) and all([combinations[i][j] for j in on_test]):
            new_gen = [j for j in range(len(combinations[i])) if combinations[i][j] and not combinations[k][j]]
            # The worst best case scenario is closest to the actual best case scenario, hence using max.
            est_cost = cost[k] + sum([d_cost[j] for j in new_gen])
            for j  in stor_abbrev:
                x_new = max(0,stor_used[j]-sum([stor_offset[j][k] for k in new_gen])) #How much of the storage discharge can be offset by new generators
                pen_dif = stor_cost[j] - .5*x_new**2*stor_avg_pen[j] #difference in discharge penalty if new generators used to offset storage use
                est_cost -= pen_dif #savings by not discharging storage assets beyond expected point
            if cost[i] != cost[i]:
                cost[i] = est_cost
            else:
                cost[i] = max([cost[i],est_cost])
    return cost



'''
Interprete the EnergyPlus format schedules to get a value at the specified time
'''
from eagers.building.basic.two_way_interp import two_way_interp

import datetime as dt

def load_sched(schedules,holidays,date,param,dd):
    profile = {}
    n_s = len(date) # number timesteps 
    h_days = find_holidays(date[0],holidays)
    ds_start,ds_end = find_daylight_savings(date[0])  
    if param is None:
        param = [x for x in schedules]
        single_sched = False
    elif len(param) ==1:
        single_sched = True
    for x in param:
        sched = schedules[x]['profile']
        profile[x]= [0 for i in range(n_s)]
        if 'timestamp' in  schedules[x]:
            print('user defined schedule')
            # profile[x] = interp_sched(sched,date) #user defined schedules defined at each time step
        else:
            for d in date:
                wd = d.weekday()
                holiday = False
                if any([x==dt.datetime(d.year,d.month,d.day) for x in h_days]):
                    holiday = True
                ds = False
                if d>ds_start and d<ds_end:
                    ds = True
                h = d.hour+d.minute/60+d.second/3600 # hour of day
                if h==0:
                    h=24
                if ds:
                    h = h+1
                    if h>24:
                        h=h-24
                hs = (d-dt.datetime(d.year,1,1)).total_seconds()/3600 #hours since start of year
                s = min([j for j in range(len(schedules[x]['seasons'])) if schedules[x]['seasons'][j]>=hs])
                inter = schedules[x]['interpolate'][s]
                ramp = schedules[x]['ramp']
                if 'AllOtherDays' in sched:
                    day = 'AllOtherDays'
                elif 'AllDays' in sched:
                    day = 'AllDays'
                elif 'ALLDays' in sched:
                    day = 'ALLDays'
                if holiday:
                    if 'Holidays' in sched:
                        day = 'Holidays'
                elif wd ==6:
                    if 'Sunday' in sched:
                        day = 'Sunday'
                    elif 'Sun' in sched:
                        day = 'Sun'
                elif wd ==5:
                    if 'Saturday' in sched:
                        day = 'Saturday'
                    elif 'Sat' in sched:
                        day = 'Sat'
                else:
                    if 'Weekday' in sched:
                        day = 'Weekday'
                    elif 'Weekdays' in sched:
                        day = 'Weekdays'                
                if dd and d.month>3:
                    day = 'SummerDesignDay'
                elif dd and d.month<=3:
                    day = 'WinterDesignDay'
                
                if inter:
                    if ramp>1e-3:
                        time,val = convert_sched(sched[day][s]['time'],sched[day][s]['val'],schedules[x]['ramp'])
                        profile[x] = two_way_interp(h,time,val) #interpolation
                    else:
                        profile[x] = two_way_interp(h,sched[day][s]['time'],sched[day][s]['val']) #interpolation
                else:#Energy plus step change at hour
                    k = 0
                    while sched[day][s]['time'][k]<h:
                        k +=1
                    profile[x] = sched[day][s]['val'][k]

    if single_sched:
        profile = profile[param[0]]       
    return profile

def find_holidays(date0,holidays):
    holiday = []
    month_names = ['january','february','march','april','may','june','july','august','september','october','november','december',]
    month_days = [31,28,31,30,31,30,31,31,30,31,30,31]
    weekday_names = ['monday','tuesday','wednesday','thursday','friday','saturday','sunday']
    for x in holidays:
        s = holidays[x]['start'].split()
        if s[0] in month_names:
            m = month_names.index(s[0])+1
            d = int(s[1])
        else:
            m = month_names.index(s[3])+1
            wd = weekday_names.index(s[1])
            dm = [t+1 for t in range(month_days[m-1]) if dt.datetime(date0.year,m,t+1).weekday()==wd]
            if s[0] == '1st':
                d = dm[0]
            elif s[0] == '2nd':
                d = dm[1]
            elif s[0] == '3rd':
                d = dm[2]
            elif s[0] == '4th':
                d = dm[3]
            elif s[0] == 'last':
                d = dm[-1]
        if date0.month>m:
            y = date0.year+1
        else:
            y = date0.year
        holiday.append(dt.datetime(y,m,d))
        
    return holiday

def convert_sched(t1,v1,r):
    if len(t1)==2:#contant value all day
        time = t1
        val = v1
    else:
        time = [t1[0]]
        val = [v1[0]]
        for t in range(1,len(t1)):
            time.append([t1[t]-min(r,(t1[t]-t1[t-1])/2),t1[t]+min(r,(t1[t]-t1[t-1])/2)])
            val.append([v1[t-1],v1[t]])
        time.append(t1[-1])
        val.append(v1[-1])
    return time,val

def find_daylight_savings(d):
    md = [dt.datetime(d.year,3,1) + dt.timedelta(days=i) for i in range(31)]
    wd = [md[x].weekday() for x in range(31)]
    sun = [i for i,x in enumerate(wd) if x==6]
    ds_start = dt.datetime(d.year,3,sun[1]) #second sunday in march

    nd = [dt.datetime(d.year,11,1) + dt.timedelta(days=i) for i in range(30)]
    wd = [nd[x].weekday() for x in range(30)]
    sun = [i for i,x in enumerate(wd) if x==6]
    ds_end = dt.datetime(d.year,11,sun[0]) #first sunday in november
    return ds_start,ds_end



def all_demands(data,subnet,t_sel):    
    squeeze, t_sel, i_sel = t_sel_indices(data,t_sel)
    # Determine the energy demands in each category at each node.
    d_n_list = list(data.keys())
    non_nodes = ["renewable", "weather", "building", "hydro"]
    for n in non_nodes:
        if n in d_n_list:
            d_n_list.remove(n)
    demand = {}
    for net in subnet['network_names']:
        demand[net] = {}
        n_list = []
        for n in range(len(subnet[net]["nodes"])):
            n_list.extend(subnet[net]["nodes"][n])

        for n in d_n_list:
            if n in n_list:
                demand[net][n] = [0 for t in t_sel]
                if 'demand' in data[n]:
                    for t in range(len(t_sel)):
                        demand[net][n][i_sel[t]] += data[n]['demand'][t_sel[t]]
    squeeze_data(demand,squeeze) 
    return demand

def add_min_building_demand(data,subnet,demand,t_sel):
    if 'building' in data and not data['building'] is None:
        for net in subnet['network_names']:
            for n in range(len(subnet[net]['nodes'])):
                if 'buildings' in subnet[net] and len(subnet[net]['buildings'][n])>0 and not subnet[net]['nodes'][n][0] in demand[net]:
                    if t_sel == None:
                        demand[net][subnet[net]['nodes'][n][0]] = [0 for t in range(len(data['timestamp']))]
                    else:
                        demand[net][subnet[net]['nodes'][n][0]] = 0
        n_b = len(data['building']['Tz_nominal'])
        b_node = {}
        for net in subnet['network_names']:
            b_node[net] = [[] for i in range(n_b)]
            for n in range(len(subnet[net]['nodes'])):
                if 'buildings' in subnet[net] and len(subnet[net]['buildings'][n])>0:
                    for b in subnet[net]['buildings'][n]:
                        b_node[net][b]=subnet[net]['nodes'][n][0]
        for i in range(n_b):   
            if t_sel == None:  
                t_sel = [t for t in range(len(data['timestamp']))]   
                for t in range(len(t_sel)):
                    if 'electrical' in demand:
                        n = b_node['electrical'][i]
                        demand['electrical'][n][t] += data['building']['E0'][t_sel[t]][i]
                    if 'district_cool' in demand:
                        n = b_node['district_cool'][i]
                        demand['district_cool'][n][t] += data['building']['c_min'][t_sel[t]][i]
                    if 'district_heat' in demand:
                        n = b_node['district_heat'][i]
                        demand['district_heat'][n][t] += data['building']['h_min'][t_sel[t]][i]
            else:
                if 'electrical' in demand:
                    demand['electrical'][b_node['electrical'][i]] += data['building']['E0'][t_sel[0]][i]
                if 'district_cool' in demand:
                    demand['district_cool'][b_node['district_cool'][i]] += data['building']['c_min'][t_sel[0]][i]
                if 'district_heat' in demand:
                    demand['district_heat'][b_node['district_heat'][i]] += data['building']['h_min'][t_sel[0]][i]


def subtract_renewable(renewable,subnet,demand):
    if 'electrical' in subnet:
        for n in range(len(subnet['electrical']['nodes'])):
            for e_name in subnet['electrical']['equipment'][n]:
                if e_name in renewable:
                    nn = subnet['electrical']['nodes'][n][0]
                    if nn in demand['electrical']:
                        demand['electrical'][nn] -= renewable[e_name]
                    else:
                        demand['electrical'][nn] = -renewable[e_name]
    if 'direct_current' in subnet:
        for n in range(len(subnet['direct_current']['nodes'])):
            for e_name in subnet['direct_current']['equipment'][n]:
                if e_name in renewable:
                    nn = subnet['direct_current']['nodes'][n][0]
                    if nn in demand['direct_current']:
                        demand['direct_current'][nn] -= renewable[e_name]
                    else:
                        demand['direct_current'][nn] = -renewable[e_name]


def t_sel_indices(data,t_sel):
    squeeze = True
    if t_sel is None:
        t_sel = [t for t in range(len(data['timestamp']))]
        squeeze = False
    i_sel = [i for i in range(len(t_sel))]
    return squeeze, t_sel, i_sel
 
def squeeze_data(demand,squeeze):
    #reduce dimension (only looking for one time step)
    if squeeze:    
        for net in demand:
            for i in demand[net]:
                demand[net][i] = demand[net][i][0]
